import site
import boto3
import json
from typing import Any, Dict, List
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.default.resources.simple import static_site
from core.constructs.output_manager import OutputTask
from core.utils import hasher


from .. import aws_client


def _create_simple_static_site(
    transaction_token: str, 
    namespace_token: str, 
    resource:  static_site.simple_static_site_model,
    output_task: OutputTask
) -> Dict:

    
    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])

    if not resource.domain_name:
        site_name = f"cdev-staticsite-{full_namespace_suffix}" 
    else:
        site_name = resource.domain_name

    output_task.update(comment=f'Creating Bucket')
    aws_client.run_client_function(
        "s3", "create_bucket", {"Bucket": site_name, "ACL": "public-read"}
    )
   
    output_task.update(comment=f'Add read object policy')
    aws_client.run_client_function(
        "s3",
        "put_bucket_policy",
        {
            "Bucket": site_name,
            "Policy": json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "PublicReadGetObject",
                            "Effect": "Allow",
                            "Principal": "*",
                            "Action": "s3:GetObject",
                            "Resource": f"arn:aws:s3:::{site_name}/*",
                        }
                    ],
                }
            ),
        },
    )

    output_task.update(comment=f'Set Website Configuration')
    aws_client.run_client_function(
        "s3",
        "put_bucket_website",
        {
            "Bucket": site_name,
            "WebsiteConfiguration": {
                "ErrorDocument": {"Key": resource.error_document},
                "IndexDocument": {"Suffix": resource.index_document},
            },
        },
    )

    ####################################
    ##### Distribution on CloudFront
    ####################################
    
    domain_name = f"{site_name}.s3-website-us-east-1.amazonaws.com"
        
    # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/using-managed-cache-policies.html#managed-cache-policies-list
    # Disabled cache policy
    cache_policy_id = '4135ea2d-6df8-44a3-9df3-4b5a84be39ad'

    args = {
        "CallerReference": str(uuid4()),
         
        "Enabled":True,
        "Origins": {
            "Quantity": 1, 
            "Items": [
                {
                    "Id": domain_name,
                    "DomainName": domain_name,
                    "CustomOriginConfig": {
                        'HTTPPort': 80,
                        'HTTPSPort': 443,
                        'OriginProtocolPolicy': 'http-only',
                        'OriginSslProtocols': {'Quantity': 3, 'Items': ['TLSv1', 'TLSv1.1', 'TLSv1.2']},
                        'OriginReadTimeout': 30,
                        'OriginKeepaliveTimeout': 5
                    },
                }
            ]
        },
        "DefaultCacheBehavior" : {
            "TargetOriginId": domain_name,
            "ViewerProtocolPolicy": 'redirect-to-https',
            "TrustedSigners": dict(Quantity=0, Enabled=False),
            "AllowedMethods": {
                'Quantity': 2, 
                'Items': ['HEAD', 'GET'], 
                'CachedMethods': {
                    'Quantity': 2, 
                    'Items': ['HEAD', 'GET']
                    }
                },
            "CachePolicyId": cache_policy_id 
        },        
        "CacheBehaviors": {'Quantity': 0},
        "Comment":"cdev generated",
        "ViewerCertificate": {
            "CloudFrontDefaultCertificate": True,
        }
    }

    if resource.domain_name:
        aliases = {
            'Aliases': {
                "Quantity": 1,
                "Items": [resource.domain_name]
            }
            
        }

        args.update(aliases)


    if resource.ssl_certificate_arn:
        ssl_args = {
            'ViewerCertificate': {
                'CloudFrontDefaultCertificate': False,
                'SSLSupportMethod': 'sni-only',
                'MinimumProtocolVersion': 'TLSv1.2_2021',
                'ACMCertificateArn': resource.ssl_certificate_arn
            }
        }

        args.update(ssl_args)


    final_args = {
        'DistributionConfig': args
    }

    output_task.update(comment=f'[blink]Creating on the Aws Cloudfront CDN. This will take a few minutes[/blink]')
    
    rv = aws_client.run_client_function('cloudfront', "create_distribution", final_args)

    cloudfront_id = rv.get("Distribution").get("Id")
    cloudfront_arn = rv.get("Distribution").get("ARN")
    cloudfront_domain = rv.get("Distribution").get("DomainName")

    aws_client.monitor_status(
        boto3.client("cloudfront").get_distribution,
        {
            "Id": cloudfront_id,
        },
        "InProgress",
        lambda x: x.get("Distribution").get("Status"),
    )


    output_info = {
        "cloud_id": cloudfront_arn,
        "bucket_name": site_name,
        "cloudfront_id": cloudfront_id,
        "cloudfront_arn": cloudfront_arn,
        "site_url": cloudfront_domain
    }

   
    #if resource.sync_folder:
    #    output_task.update(comment=f'Sync Files')
    #    file_syncer = sync_files()
    #    file_syncer.command(
    #        **{"resource_name": resource.name, "dir": resource.content_folder}
    #    )


    return output_info


def _update_simple_static_site(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: static_site.simple_static_site_model,
    new_resource: static_site.simple_static_site_model,
    previous_output: Dict,
    output_task: OutputTask
) -> bool:

    if not previous_resource.domain_name == new_resource.domain_name:
        raise Exception("Can not update 'domain_name' via an in place update. You must destroy this version and make a new static site resource.")


    if not previous_resource.ssl_certificate_arn == new_resource.ssl_certificate_arn:
        raise Exception("Can not update 'ssl_certificate_arn' via an in place update. You must destroy this version and make a new static site resource.")



    bucket_name = previous_output.get('bucket_name')

    if not previous_resource.index_document == new_resource.index_document:
        aws_client.run_client_function(
            "s3",
            "put_bucket_website",
            {
                "Bucket": bucket_name,
                "WebsiteConfiguration": {
                    "ErrorDocument": {"Key": new_resource.error_document},
                    "IndexDocument": {"Suffix": new_resource.index_document},
                },
            },
        )



    if not previous_resource.error_document == new_resource.error_document:
        aws_client.run_client_function(
            "s3",
            "put_bucket_website",
            {
                "Bucket": bucket_name,
                "WebsiteConfiguration": {
                    "ErrorDocument": {"Key": new_resource.error_document},
                    "IndexDocument": {"Suffix": new_resource.index_document},
                },
            },
        )


    if not previous_resource.sync_folder == new_resource.sync_folder:
        # If there has been a change in sync from false to true then sync the new directory
        #if new_resource.sync_folder:
        #    output_task.update(comment=f'Syncing Files')
        #    
        #    file_syncer = sync_files()
        #    file_syncer.command(
        #        **{
        #            "resource_name": new_resource.name,
        #            "dir": new_resource.content_folder,
        #        }
        #    )
            
        #else:
        #    output_task.update(comment=f'Disbled Sync Files')
        pass


    #elif not previous_resource.content_folder == new_resource.content_folder:
    #    output_task.update(comment=f'Sync Files')
#
    #    # If there was not change to sync but there is a change to the content folder. Clear the bucket and sync new folder
    #    s3 = boto3.resource("s3")
    #    bucket = s3.Bucket(site_name)
    #    bucket.object_versions.delete()
    #   
    #    file_syncer = sync_files()
    #    file_syncer.command(
    #        **{"resource_name": new_resource.name, "dir": new_resource.content_folder}
    #    )
    #    

    # No changes affect the actual cloud output
    return previous_output

def _remove_simple_static_site(
    transaction_token: str, 
    previous_output: Dict, 
    output_task: OutputTask
):
    previous_cloudfront_id = previous_output.get('cloudfront_id')

    all_info = aws_client.run_client_function("cloudfront", "get_distribution", 
        {
            "Id": previous_cloudfront_id
        }
    )


    current_distribution_info = all_info.get('Distribution').get('DistributionConfig')
    e_tag = all_info.get("ETag")

    current_distribution_info['Enabled'] = False

    final_args = {
        "DistributionConfig": current_distribution_info,
        "Id": previous_cloudfront_id,
        "IfMatch": e_tag

    }

    # Disable the distribution
    new_info =  aws_client.run_client_function("cloudfront", "update_distribution", final_args)
    new_etag = new_info.get("ETag")

    output_task.update(comment='[blink]Disabling site on Aws Cloudfront CDN.This will take a few minutes[/blink]')
    aws_client.monitor_status(
        boto3.client("cloudfront").get_distribution,
        {
            "Id": previous_cloudfront_id
        },
        "InProgress",
        lambda x: x.get("Distribution").get("Status"),
    )


    output_task.update(comment='Deleting site from Aws Cloudfront CDN')
    aws_client.run_client_function("cloudfront", "delete_distribution",
        {
            "Id": previous_cloudfront_id,
            "IfMatch": new_etag
        }
    )

    bucket_name = previous_output.get('bucket_name')

    s3 = boto3.resource("s3")
    bucket = s3.Bucket(bucket_name)
    output_task.update(comment='Deleting all items in the bucket')
    bucket.object_versions.delete()

    output_task.update(comment='Deleting bucket')
    aws_client.run_client_function(
        "s3", "delete_bucket", {"Bucket": bucket_name}
    )



def handle_simple_static_site_deployment(
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_static_site(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:
        return _update_simple_static_site(
            transaction_token,
            namespace_token,
            static_site.simple_static_site_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )

    elif resource_diff.action_type == Resource_Change_Type.DELETE:
        _remove_simple_static_site(
            transaction_token,
            previous_output,
            output_task
        )

        return {}



import json
import os
from typing import Callable, List, Dict, Optional, Any, Tuple, Union
from pydantic.main import BaseModel

from pydantic.types import DirectoryPath, FilePath

from core.constructs.backend import Backend, Backend_Configuration, load_backend
from core.constructs.mapper import CloudMapper
from core.constructs.components import Component, ComponentModel
from core.constructs.workspace import (
    Workspace_State,
    Workspace_Info,
    Workspace,
    WorkspaceManager,
    wrap_phase,
)
from core.constructs.cloud_output import Cloud_Output, cloud_output_dynamic_model, evaluate_dynamic_output, cloud_output_model

from ..utils import file_manager, module_loader


DEFAULT_COMMAND_LOCATIONS = [
    'core.default.commands'
]

ROOT_FOLDER_NAME = ".cdev"
WORKSPACE_FILE_NAME = "workspace_info.json"

class local_workspace_configuration(BaseModel):
    initialization_module: str
    backend_configuration: Backend_Configuration
    resource_state_uuid: Optional[str]


class local_workspace(Workspace):
    """
    A singleton that encapsulates the configuration of a workspace that is implemented on the local filesystem. The singleton can be accessed during different
    parts of the lifecycle of a cdev core command execution. When this singleton is created, it is registered as the global workspace for that execution.
    """

    _instance = None

    _COMMANDS = DEFAULT_COMMAND_LOCATIONS
    _MAPPERS = []
    _COMPONENTS = []

    _resource_state_uuid: str = None

    _backend: Backend = None

    _state: Workspace_State = Workspace_State.UNINITIALIZED

    _output: List[Tuple[str, str, cloud_output_model]] = []

    _current_component: str = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Workspace, cls).__new__(cls)
            # Put any initialization here.

            cls._instance.set_state(Workspace_State.UNINITIALIZED)

            # Load the backend
            cls._instance._backend = None

            cls._instance._COMMANDS = DEFAULT_COMMAND_LOCATIONS
            cls._instance._MAPPERS = []
            cls._instance._COMPONENTS = []
            cls._instance._resource_state_uuid = None

            Workspace.set_global_instance(cls._instance)

        return cls._instance

    @classmethod
    def terminate_singleton(cls):
        cls._instance = None

    def initialize_workspace(
        self, workspace_configuration_dict: local_workspace_configuration
    ):

        # It is the responsibility of the higher up callers to set the correct states for initialization.
        # This allows the flexibility of higher up frameworks injecting initialization steps into the process.
        workspace_configuration = local_workspace_configuration(
            **workspace_configuration_dict
        )
        
        try:
            backend_config = workspace_configuration.backend_configuration
            self.set_backend(load_backend(backend_config))
        except Exception as e:
            print(f"Could not load the load backend")
            raise e

        module_loader.import_module(workspace_configuration.initialization_module, denote_output=True)
        
        if workspace_configuration.resource_state_uuid:
            if not workspace_configuration.resource_state_uuid in set(
                [x.uuid for x in self._backend.get_top_level_resource_states()]
            ):
                raise Exception(
                    f"{workspace_configuration.resource_state_uuid} not in loaded backend ({self._backend.get_top_level_resource_states()})"
                )

        self.set_resource_state_uuid(workspace_configuration.resource_state_uuid)
        
        

    def destroy_workspace(self):
        Workspace.destroy_workspace(self)
        local_workspace.terminate_singleton()

    ################
    ##### State
    ################
    def set_state(self, value: Workspace_State):
        self._state = value

    def get_state(self) -> Workspace_State:
        return self._state


    @wrap_phase([Workspace_State.EXECUTING_FRONTEND])
    def generate_current_state(self) -> List[ComponentModel]:
        """
        Execute the components of this workspace to generate the current desired state of the resources.

        Returns:
            Current State (List[ComponentModel]): The current state generated by the components.
        """

        rv = []
        components: List[Component] = self.get_components()

        for component in components:
            self._current_component = component.name
            rv.append(component.render())

        return rv


    #######################
    ##### Display Output
    #######################
    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_FRONTEND, Workspace_State.EXECUTING_BACKEND])
    def display_output(self, tag: str, output: Cloud_Output):
        self._output.append((tag, self._current_component, output.render()))


    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_BACKEND])
    def render_outputs(self) -> List[Tuple[str, Any]]:
        """Render the output associated with the Workspace

        Returns:
            List[Tuple[str, Any]]: The List of outputs with their associated tag
        
        """
        rv = []


        if not self._current_component:
            raise Exception

        for tag, component, cloud_output in self._output:
            resolved_value = self.get_backend().get_cloud_output_value_by_name(
                self.get_resource_state_uuid(),
                component,
                cloud_output.ruuid,
                cloud_output.name,
                cloud_output.key
            )

            if isinstance(cloud_output, cloud_output_dynamic_model):
                resolved_value = evaluate_dynamic_output(
                    resolved_value,
                    cloud_output
                )

            rv.append((tag, resolved_value))

        return rv

    #################
    ##### Mappers
    #################
    @wrap_phase([Workspace_State.INITIALIZING])
    def add_mapper(self, mapper: CloudMapper) -> None:
        if not isinstance(mapper, CloudMapper):
            # TODO Throw error
            print(f"BAD CLOUD MAPPER {mapper}")
            return

        self._MAPPERS.append(mapper)

    @wrap_phase([Workspace_State.INITIALIZING])
    def add_mappers(self, mappers: List[CloudMapper]) -> None:
        for mapper in mappers:
            self.add_mapper(mapper)

        self._MAPPERS.append(mapper)

    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_BACKEND])
    def get_mappers(self) -> List[CloudMapper]:
        return self._MAPPERS

    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_BACKEND])
    def get_mapper_namespace(self) -> Dict:
        rv = {}
        mappers: List[CloudMapper] = self.get_mappers()

        for mapper in mappers:
            for namespace in mapper.get_namespaces():
                rv[namespace] = mapper

        return rv

    #################
    ##### Commands
    #################
    @wrap_phase([Workspace_State.INITIALIZING])
    def add_command(self, command_location: str):
        self._COMMANDS.append(command_location)

    @wrap_phase([Workspace_State.INITIALIZING])
    def add_commands(self, command_locations: List[str]):
        for command_location in command_locations:
            self.add_command(command_location)

    @wrap_phase([Workspace_State.INITIALIZED])
    def get_commands(self) -> List[str]:
        return self._COMMANDS

    #################
    ##### Components
    #################
    @wrap_phase([Workspace_State.INITIALIZING])
    def add_component(self, component: Component):
        self._COMPONENTS.append(component)

    @wrap_phase([Workspace_State.INITIALIZING])
    def add_components(self, components: List[Component]):
        for component in components:
            self.add_component(component)

    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_FRONTEND])
    def get_components(self) -> List[Component]:
        return self._COMPONENTS

    #################
    ##### Backend
    #################
    @wrap_phase([Workspace_State.INITIALIZING])
    def set_backend(self, backend: Backend):
        if not isinstance(backend, Backend):
            raise Exception("Not a backend object")

        self._backend = backend

    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_FRONTEND, Workspace_State.EXECUTING_BACKEND])
    def get_backend(self) -> Backend:
        return self._backend

    @wrap_phase([Workspace_State.INITIALIZING])
    def set_resource_state_uuid(self, resource_state_uuid: str):
        self._resource_state_uuid = resource_state_uuid

    @wrap_phase([Workspace_State.INITIALIZED, Workspace_State.EXECUTING_FRONTEND,  Workspace_State.EXECUTING_BACKEND])
    def get_resource_state_uuid(self) -> str:
        return self._resource_state_uuid


class local_workspace_manager(WorkspaceManager):
    def __init__(
        self,
        base_dir: DirectoryPath,
        workspace_dir: str = None,
        workspace_filename: str = None,
    ) -> None:
        self.base_dir = base_dir
        self.workspace_dir = (
            workspace_dir if workspace_dir else ROOT_FOLDER_NAME
        )
        self.workspace_filename = (
            workspace_filename
            if workspace_filename
            else WORKSPACE_FILE_NAME
        )

    def create_new_workspace(self, workspace_info: Workspace_Info):
        """
        Create a new workspace based on the information provided.

        Args:
            workspace_info (Workspace_Info): information about the backend configuration

        Raises:
            WorkSpaceAlreadyCreated
        """
        base_cdev_dir = os.path.join(self.base_dir, self.workspace_dir)
        if not os.path.isdir(base_cdev_dir):
            os.mkdir(base_cdev_dir)

        file_manager.safe_json_write(
            workspace_info.dict(), os.path.join(base_cdev_dir, self.workspace_filename)
        )

    def check_if_workspace_exists(self) -> bool:
        return os.path.isfile(
            os.path.join(self.base_dir, self.workspace_dir, self.workspace_filename)
        )

    def load_workspace_configuration(self) -> Workspace_Info:
        file_location = os.path.join(
            self.base_dir, self.workspace_dir, self.workspace_filename
        )

        with open(file_location, "r") as fh:
            raw_data = json.load(fh)

        try:
            rv = Workspace_Info(**raw_data)

            return rv
        except Exception as e:
            print(e)
            raise e



"""Utilities to help work with the Resource Dependency Graph.

One of the core components of the Cdev Core framework is the Resource Dependency Graph. This graph is implemented
using the `networkx` packages as it provides helpful utilities for working with graph data structures. 
"""

from concurrent.futures import ThreadPoolExecutor, Future
from enum import Enum

from networkx.classes.digraph import DiGraph
from networkx.classes.reportviews import NodeView


from typing import Callable, Dict, List, Set, Tuple
from time import sleep

from core.constructs.resource import Resource_Change_Type, Resource_Reference_Change_Type, ResourceModel, Resource_Difference, Resource_Reference_Difference
from core.constructs.cloud_output import OutputType, cloud_output_model
from core.constructs.components import Component_Change_Type, Component_Difference
from core.constructs.output_manager import OutputManager, OutputTask
from core.constructs.models import frozendict

deliminator = '+'

def find_parents(resource: ResourceModel) -> Tuple[List, List]:
    """Find any parents resources via any linked Cloud Output Models

    If a resource constains the cloud output of another resource, the relationship
    needs to be identified as a parent-child relationship with the Resource Dependency
    Graph.

    Args:
        resource (ResourceModel): Resource to look for any parents with

    Returns:
        Tuple[
            parent_resources: List[Tuple[ruuid, name]]
            parent_resources: List[Tuple[ruuid, name]]
        ]
    """
    resource_as_obj = resource.dict()
    
    cloud_outputs = find_cloud_output(resource_as_obj)

    parent_resources = [(x.ruuid,x.name) for x in cloud_outputs if x.type == OutputType.RESOURCE ]
    parent_references = [(x.ruuid,x.name) for x in cloud_outputs if x.type == OutputType.REFERENCE]
    

    return parent_resources, parent_references


def find_cloud_output(obj: dict) -> List[cloud_output_model]:

    rv = _recursive_find_output(obj)

    return list(set(rv))


def _recursive_find_output(obj) -> List[cloud_output_model]:
    rv = []

    if isinstance(obj, frozendict) or isinstance(obj, dict): 
        
        if "id" in obj and obj.get("id") == 'cdev_cloud_output':
            rv.append(cloud_output_model(**obj))
        else:
            for k,v in obj.items():
                if isinstance(v, frozendict) or isinstance(v, dict):
                    if "id" in v and v.get("id") == 'cdev_cloud_output':
                        rv.append(cloud_output_model(**v))
                    else:
                        rv.extend(_recursive_find_output(v))

                elif isinstance(v, frozenset) or isinstance(v, list):
                    all_items_rendered = [_recursive_find_output(x) for x in v]

                    for item in all_items_rendered:
                        rv.extend(item)

        return rv

    elif isinstance(obj, frozenset) or isinstance(obj, list):
        all_items_rendered = [_recursive_find_output(x) for x in obj]
                
        for item in all_items_rendered:
            rv.extend(item)

    return rv


def generate_sorted_resources(differences: Tuple[List[Component_Difference], List[Resource_Difference], List[Resource_Reference_Difference]]) -> DiGraph:
    # nx graphs work on the element level by using the __hash__ of objects added to the graph, so all the elements added to the graph should be a pydantic Model with
    # the 'frozen' feature set
    change_dag = DiGraph()


    component_differences = differences[0]
    resource_differences = differences[1]
    reference_differences = differences[2]
    

    component_ids = {_create_component_id(x.new_name):x for x in component_differences if x.action_type == Component_Change_Type.CREATE or x.action_type == Component_Change_Type.UPDATE_NAME or x.action_type == Component_Change_Type.UPDATE_IDENTITY}
    component_ids.update({_create_component_id(x.previous_name):x for x in component_differences if x.action_type == Component_Change_Type.DELETE})

    
    resource_ids = {_create_resource_id(x.component_name, x.new_resource.ruuid, x.new_resource.name):x for x in resource_differences if x.action_type == Resource_Change_Type.UPDATE_IDENTITY or x.action_type == Resource_Change_Type.CREATE or x.action_type == Resource_Change_Type.UPDATE_NAME}
    resource_ids.update({_create_resource_id(x.component_name, x.previous_resource.ruuid, x.previous_resource.name):x for x in resource_differences if x.action_type == Resource_Change_Type.DELETE})

    
    reference_ids = {_create_reference_id(x.originating_component_name, x.resource_reference.component_name, x.resource_reference.ruuid, x.resource_reference.name):x for x in reference_differences}

    
    for _, component in component_ids.items():
        change_dag.add_node(component)


    for _, resource in resource_ids.items():
        change_dag.add_node(resource)

        component_id = _create_component_id(resource.component_name)

        if component_id in component_ids:
            if resource.action_type == Resource_Change_Type.DELETE and component_ids.get(component_id).action_type == Component_Change_Type.DELETE:
                # Since these are both deletes, it should happen in the reverse order
                change_dag.add_edge(resource,component_ids.get(component_id))

            else:
                change_dag.add_edge(component_ids.get(component_id), resource)
        else:
            raise Exception(f"There should always be a change in a component for a resource change {resource}")

        parent_resources, parent_references = find_parents(resource.new_resource) if not resource.action_type == Resource_Change_Type.DELETE else find_parents(resource.previous_resource)

        if not parent_resources and not parent_references:
            continue


        for parent_ruuid, parent_name in parent_resources:
            # Parent resources must be in the same component
            
            # Cloud outputs of a resource will always be from the same component
            parent_resource_id = _create_resource_id(resource.component_name, parent_ruuid, parent_name)

            if parent_resource_id in resource_ids:
                if resource.action_type == Resource_Change_Type.DELETE:
                    # Since this is a delete, it should happen in the reverse order
                    change_dag.add_edge(resource, resource_ids.get(parent_resource_id))

                else:
                    # Make this resource change a child of the parent resource change
                    change_dag.add_edge(resource_ids.get(parent_resource_id), resource)

        for parent_ruuid, parent_name in parent_references:
            parent_reference_id = _create_reference_id(resource.component_name, resource.component_name, parent_ruuid, parent_name)

            if parent_reference_id in reference_ids:
                if resource.action_type == Resource_Change_Type.DELETE:
                    # Since this is a delete, it should happen in the reverse order
                    change_dag.add_edge(resource, resource_ids.get(parent_reference_id))

                else:
                    change_dag.add_edge(resource_ids.get(parent_reference_id), resource)

    for _, reference in reference_ids.items():
        change_dag.add_node(reference)

        # Reference changes should be a child to an Update Identity component diff
        originating_component_id = _create_component_id(reference.originating_component_name)

        if originating_component_id in component_ids:
            change_dag.add_edge(component_ids.get(originating_component_id), reference)

        else:
            raise Exception(f"There should always be a change in a component for a reference change {resource}")


        # Reference change should also be a child to any change to the referenced resource
        resource_id = _create_resource_id(reference.resource_reference.component_name, reference.resource_reference.ruuid, reference.resource_reference.name)
        if resource_id in resource_ids:
            # If there is also a change to the component this reference is apart of then the changes need to be ordered
            if reference.action_type == Resource_Reference_Change_Type.CREATE:
                if resource_ids.get(resource_id).action_type == Resource_Change_Type.DELETE:
                    # We can not create a reference if that resource is currently being destroyed.
                    raise Exception(f"Trying to reference a resource that is being deleted {reference} -> {resource_id}")

                # Since the reference is being created, we should perform the operation after the changes to the underlying resource 
                change_dag.add_edge(resource_ids.get(resource_id), reference)

            else:
                change_dag.add_edge(reference, resource_ids.get(resource_id))

        else:
            # Since the original resource is not in the change set, it should be checked to be sure that it is accessible from the backend
            pass
            

    return change_dag


def _create_component_id(component_name: str ) -> str:
    # Component Differences will be identified by the key component<deliminator><component_name>
    return f'component{deliminator}{component_name}'


def _create_resource_id(component_name: str, ruuid: str, name: str) -> str:
    # Resource Differences will be identified by the key resource<deliminator><ruuid><deliminator><name>
    return f'resource{deliminator}{component_name}{deliminator}{ruuid}{deliminator}{name}'


def _create_reference_id(originating_component_name: str, component_name: str, ruuid: str, name: str) -> str:
    # Reference Differences will be identified by the key reference<deliminator><ruuid><deliminator><name>
    return f'reference{deliminator}{originating_component_name}{deliminator}{component_name}{deliminator}{ruuid}{deliminator}{name}'




class node_state(str, Enum):
    UNPROCESSED = "UNPROCESSED"
    PROCESSING = "PROCESSING"
    PROCESSED = "PROCESSED"
    ERROR = "ERROR"
    PARENT_ERROR = "PARENT_ERROR"


def topological_iteration(dag: DiGraph, process: Callable[[NodeView], None], failed_parent_handler: Callable[[NodeView], None]=None, thread_count: int = 1, interval: float = .3, pass_through_exceptions: bool = False):
    """Execute the `process` over a DAG in a topologically constrained way. 
    
    This means that the `process` will not be executed on a node until the `process` has been executed on all parents of that node. 
    
    If the `process` throws an Exception when executing on a Node, then any child nodes will not be executed (if provided, the 
    failed_parent_handler will be executed on the child nodes). 
    
    This iteration supports multiple threads via the `thread_count` param. Multiple threads can speed up the total iteration time if the `process` 
    is not CPU bound and there are non-dependant paths through the DAG. Note that the `process` provided should be thread safe when using multiple threads.

    Args:
        dag (DiGraph): The graph to execute over.
        process (Callable[[NodeView], None]): The function to call on each Node.
        failed_parent_handler (Callable[[NodeView], None], optional): A function to call on Nodes that do not execute because a parent failed.
        thread_count (int, optional): [description]. Defaults to 1 
        interval (float, optional): Interval (in seconds) to poll threads for completion. Defaults to .3
    """
    
    all_children: Set[NodeView] = set(x[1] for x in dag.edges)
    all_nodes: Set[NodeView] = set(x for x in dag.nodes())

    # starting nodes are those that have no parents
    starting_nodes = all_nodes - all_children
    
    nodes_to_process: list[NodeView] = []
    nodes_to_process.extend(starting_nodes)

    _processing_future_to_resource: Dict[Future, NodeView] = {}

    # Keep track of the state of all nodes to determine when the entire DAG has been traversed
    _node_to_state: Dict[NodeView,node_state] = {x:node_state.UNPROCESSED for x in all_nodes}

    executor = ThreadPoolExecutor(thread_count)
    
    # While any node is unprocessed or still processing
    while any(_node_to_state.get(x) == node_state.UNPROCESSED or _node_to_state.get(x) == node_state.PROCESSING for x in all_nodes):
        
        # Pull any ready nodes to be processed and add them to the thread pool
        for _ in range(0, len(nodes_to_process)):
            node_to_process = nodes_to_process.pop(0)
            future = executor.submit(process, node_to_process)

            _processing_future_to_resource[future] = node_to_process
            _node_to_state[node_to_process] = node_state.PROCESSING

        
        # Check if any of the futures are finished and then make decisions about their children from the rv of the future
        for fut, node in _processing_future_to_resource.copy().items():
            # Note python throws runtime error if you change a dict size while iterating so use a copy for now.
            # TODO: Check if there is a more optimized way to do this

            if not fut.done():
                # Still processing
                continue


            try:
                result = fut.result()

                # No exceptions raised so process completed correctly
                _node_to_state[node] = node_state.PROCESSED

                children = dag.successors(node)

                for child in children:
                    # If any of the parents of the child have not been processed, than the child node is not ready to be processed
                    if any(not _node_to_state.get(x) == node_state.PROCESSED for x in dag.predecessors(child)):
                        continue
                    
                    nodes_to_process.append(child)

            except Exception as e:
                # Since this returned a error need to mark all children as unable to deploy
                _node_to_state[node] = node_state.ERROR
                
                  
                # mark an descdents of this node as unable to process
                _recursively_mark_parent_failure(_node_to_state, dag, node, handler=failed_parent_handler)

                if pass_through_exceptions:
                    raise e 
                         

            # Remove the future from the dictionary
            _processing_future_to_resource.pop(fut)

        sleep(interval)


    executor.shutdown()


def _recursively_mark_parent_failure(_node_to_state: Dict[NodeView, node_state], dag: DiGraph, parent_node: NodeView, handler: Callable[[NodeView, OutputTask], None]=None, pass_through_exceptions: bool = False) -> None:

    if parent_node not in _node_to_state:
        raise Exception(f"trying to mark node ({parent_node}) but cant not find it in given dict")

    
    children = dag.successors(parent_node)

    for child in children:

        if child not in _node_to_state:
            raise Exception(f"trying to mark node ({child}) but cant not find it in given dict")
        
        _node_to_state[child] = node_state.PARENT_ERROR

        if handler:
            # Call a handler that does extra busines logic for items that fail because of their parent
            try:
                handler(child)

            except Exception as e:
                #handler threw exception but we should continue
                
                if pass_through_exceptions:
                    raise e 


        _recursively_mark_parent_failure(_node_to_state, dag, child)




from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on the simple static site.

sync: Sync the contents of a given folder to a deployed site

"""



"""Module that contains functionality to manage the understanding of packages within this `Workspace`
"""

import os
from pathlib import Path
import pkg_resources
from pydantic.types import DirectoryPath, FilePath
from sortedcontainers.sorteddict import SortedDict
import sys
from typing import List, Set, Dict, Tuple, Union, Optional

from serverless_parser import parser as cdev_parser

from . import docker_package_builder
from .utils import PackageTypes, ModulePackagingInfo, environment_to_architecture_suffix, Architecture

from core.utils.platforms import lambda_python_environment, get_current_closest_platform
from core.utils.logger import log

# Keep cache of already seen package names
PACKAGE_CACHE = {}

# There isn't a great runtime way of identifying python standard libraries (non C builtin libraries packaged with python).
# So I scraped the information from the python documentation website using the ./scripts/list-python-builtins
STANDARD_LIBRARY_FILES = ["3_7", "3_8", "3_9"]

# Build a dict of pkg name to pip package obj so you don't have to loop over all the packages when doing a look up
# We need to include both a the project name and 'toplevel' name. This is because a project can have a python unsafe name
# so it is allowed to have a different name that is used when importing the package in an actual python file. Both need to
# be include because, when the package is reference directly in a file it will have the top name, but when it is referenced
# as the dependency of another project it will be a the project name.
MOD_NAME_TO_PRJ_OBJ: Dict[str, pkg_resources.Distribution] = {}

PRJ_NAME_TO_TOP_LEVEL_MODULES: Dict[str, List[str]] = {}

PLATFORM_DEPENDENT_PROJECTS = set()

_already_checked_cache = set()

download_package_location = None

target_platform = get_current_closest_platform()
CURRENT_PLATFORM = get_current_closest_platform()

#_current_working_dir = CDEV_SETTINGS.get("CURRENT_PARSING_DIR")
# TODO: Make this a setting
_current_working_dir = os.getcwd()

environment_to_python_std = {
    lambda_python_environment.py37: "3_7",
    lambda_python_environment.py38_x86_64: "3_8",
    lambda_python_environment.py38_arm64: "3_8",
    lambda_python_environment.py39_x86_64: "3_9",
    lambda_python_environment.py39_arm64: "3_9",
    lambda_python_environment.py3_x86_64: "3_9",
    lambda_python_environment.py3_arm64: "3_9",
}


def _is_platform_compatible(tags: List[str]) -> bool:
    # https://packaging.python.org/specifications/platform-compatibility-tags/
    # PEP 600
    interpreter_tag = tags[0]
    platform_tag = tags[-1]

    if not (interpreter_tag[:2] == "py" or interpreter_tag[:2] == "cp"):
        # Not a cpython or generic python package
        raise Exception

    if platform_tag == "any":
        return True

    if "win32" in platform_tag:
        return False

    elif "macosx" in platform_tag:
        return False

    else:
        #Linux specific package
        return False


for project_obj in pkg_resources.working_set:
    # We need to compute some information about the available modules in the current environment. This will help to provide information about
    # the modules when handlers reference them.

    # Compute:
    # MOD_NAME_TO_PRJ_OBJ -> Dict from the module name to the project object that contains the metadata about the project the module is from
    # PRJ_NAME_TO_TOP_LEVEL_MODULES -> Dict from the project name to all of its top level modules which are used when including the project as a dependency
    # INCOMPATIBLE_PROJECTS -> Projects that are not platform agnostic and would therefore need to be redownloaded for the correct deployment architecture 

    # For information on this object check
    # https://setuptools.pypa.io/en/latest/pkg_resources.html#distribution-objects
    # Distribution Object

    # find the dist info directory that will contain metadata about the package
    dist_dir_location = os.path.join(
        project_obj.location,
        f"{project_obj.project_name.replace('-', '_')}-{project_obj.parsed_version}.dist-info",
    )
    toplevel_file_location = os.path.join(dist_dir_location, "top_level.txt")
    wheel_info = os.path.join(dist_dir_location, "WHEEL")
    
    if not os.path.isdir(dist_dir_location):
        continue

    if not os.path.isfile(toplevel_file_location):
        # If not top level file is present, then assume the only top level module available is the project name
        
        MOD_NAME_TO_PRJ_OBJ[project_obj.project_name.replace('-','_')] = project_obj
        PRJ_NAME_TO_TOP_LEVEL_MODULES[project_obj.project_name] = [
            project_obj.project_name
        ]
        continue

    with open(toplevel_file_location) as fh:
        top_level_mod_names = fh.readlines()

        for top_level_mod_name in top_level_mod_names:
            MOD_NAME_TO_PRJ_OBJ[top_level_mod_name.strip()] = project_obj

        potential_modules = [x.strip() for x in top_level_mod_names]
        actual_modules = []
        for potential_module in potential_modules:
            # The module could be either a folder (normal case) or a single python file (ex: 'six' package)
            # If it can not be found as either than there is an issue
            potential_dir = os.path.join(project_obj.location, potential_module)
            potential_file = os.path.join(
                project_obj.location, potential_module + ".py"
            )

            if not os.path.isdir(potential_dir) and not os.path.isfile(potential_file):
                # print(f"Could not find module {potential_module} at {dist_dir_location}")
                continue

            actual_modules.append(potential_module)

        PRJ_NAME_TO_TOP_LEVEL_MODULES[project_obj.project_name] = actual_modules

    with open(wheel_info) as fh:
        lines = fh.readlines()

        # https://www.python.org/dev/peps/pep-0425/
        # if it is not pure it should only have one tag
        # We are going ot check the tags of the package to make sure it is compatible with the target deployment platform
        tags = (
            [x.split(":")[1] for x in lines if x.split(":")[0] == "Tag"][0]
            .strip()
            .split("-")
        )

        if not _is_platform_compatible(tags):
            PLATFORM_DEPENDENT_PROJECTS.add(project_obj.project_name)

######################
##### Helper Functions
######################

def _load_aws_packages(version="3_6"):
    """Returns the given built in third party packages for a given platform

    Args:
        version (str, optional): [description]. Defaults to "3_6".

    Returns:
        set[str]: modules
    """
    return set(["boto3", "botocore"])


def _load_mod_to_prj():
    return MOD_NAME_TO_PRJ_OBJ


def _load_standard_library_information(version="3_6"):
    FILE_LOC = os.path.join(
        os.path.dirname(__file__), "standard_library_names", f"python_{version}"
    )

    if not os.path.isfile(FILE_LOC):
        # TODO throw error
        raise FileNotFoundError

    with open(FILE_LOC) as fh:
        return set(fh.read().splitlines())

def _clear_already_checked_cache(starting_location: str):
    # Add the starting file of the recursive calls
    global _already_checked_cache

    _already_checked_cache = set()

    _already_checked_cache.add(starting_location)

##########################
##### Main Entry Point
##########################

def get_top_level_module_info(
    modules: List[str], start_location: FilePath, target_platform_param: lambda_python_environment, download_package_location_param: DirectoryPath = None,
) -> Dict[str, ModulePackagingInfo]:
    """Given a set of modules, find the information needed to package those modules. 

    Args:
        modules (List[str]): Module names used to by a handler
        target_platform_param (lambda_python_environment): environment the function will be deployed on, which can affect what version of
        a 3rd party module is used.
        download_package_location_param (DirectoryPath): Base Directory that any downloaded package from Docker will be stored.

    Returns:
        module_infos (SortedDict[str, ModulePackagingInfo]): Dict (Sorted by module name) of the module information
        objects


    A handler file will have a set top level modules that are defined at the top of the file. We need to determine the module information
    about these and then return them in a way that allows them to be packaged efficiently. 


    """
    
    # Set the global value for the recursive function to use
    global download_package_location
    download_package_location = download_package_location_param

    global target_platform
    target_platform = target_platform_param

    log.debug('Getting Module Packaging Info for modules %s for target platform %s', modules, target_platform)
    
    all_packages = {}

    # we want to keep a cache of files we have already checked to speed up the process
    # clear this cache between iterations to make sure no files have changed
    _clear_already_checked_cache(start_location)

    for module_name in modules:
        all_packages[module_name] = _get_module_info(module_name, start_location)

    return SortedDict(all_packages)


def _get_module_info(
    module_name: str, original_file_location: str
) -> ModulePackagingInfo:
    """Create the information needed to package this dependency with a parsed function. 
    
    Args:
        module_name (str): The module name to look up. It can also be a relative name (begins with '.')
        original_file_location (str): Since the module name can be a relative import it needs to have the location of the
        starting location to help resolve the reference

    Returns
        info (ModulePackagingInfo): information for the packages to package

    We use the recursive method because we must compute the package information for the dependencies of this dependency. 
   
    """
    log.debug('Top Getting Module Packaging Info for module %s', module_name)
    # Note the package cache is implemented at the recursive level so that recursive calls can benefit from the cache also
    info = _recursive_create_module_package_info(module_name, original_file_location)

    log.debug('Top Recieve Module Packaging Info %s for module %s', info, module_name)

    return info





def _recursive_create_module_package_info(
    module_name: str, original_file_location: str
) -> ModulePackagingInfo:
    """Recursively create a module information object by finding out what type of module this is, and then recursively creating
    any dependant modules.

    Args:
        module_name (str): This is the top level module name used by the handler (can be a relative module)
        original_file_location: Location to use as start point if a local module is given

    Returns:
        module_info (ModulePackagingInfo): Information object for the given module

    """

    global download_package_location
    global target_platform

    if (module_name, original_file_location) in PACKAGE_CACHE:
        # Look in the cache if there is already information about this module to speed up the process
        log.debug('Package CACHE HIT for  (%s,%s)', module_name, original_file_location)
        return PACKAGE_CACHE.get((module_name, original_file_location))

    if (
        not module_name in sys.modules
        and not module_name in MOD_NAME_TO_PRJ_OBJ
        and not module_name[0] == "."
    ):
        # The module name is not in the available system modules and also not a relative module
        raise Exception(f"Bad module name {module_name}; Not a sys module, relatively referenced package, or pip installed package")

    else:
        standard_lib_info = _load_standard_library_information(environment_to_python_std.get(target_platform))
        aws_packages = _load_aws_packages(environment_to_python_std.get(target_platform))
        pip_packages = _load_mod_to_prj()

        if module_name in standard_lib_info:
            # Module is from the standard library
            log.debug('Standard Library: %s', module_name)
            rv = ModulePackagingInfo(
                **{
                    "module_name": module_name,
                    "type": PackageTypes.STANDARDLIB,
                    "version_id": None,
                    "fp": None,
                }
            )

        elif module_name in aws_packages:
            # Module is part of the default libraries available in the aws lambda python environment
            log.debug('Library included with runtime environment: %s', module_name)
            rv = ModulePackagingInfo(
                **{
                    "module_name": module_name,
                    "type": PackageTypes.AWSINCLUDED,
                    "version_id": None,
                    "fp": None,
                }
            )

        elif module_name in pip_packages:
            # Module was installed with a package manager and therefor contains additional metadata that can be used to find the dependencies
            log.debug('Pip package: %s', module_name)
            tmp_distribution_obj = pip_packages.get(module_name)

            project_name = tmp_distribution_obj.project_name

            module_arch = Architecture.ANY

            if project_name in PLATFORM_DEPENDENT_PROJECTS:
                log.debug('Pip package %s is not a universal package', module_name)
                # Some of the projects that can be installed are platform dependant and the users environment might not match the aws lambda
                # environment. So, we need to use docker to pull the compatible version of the library then use that in the final archive
                
                if CURRENT_PLATFORM and (CURRENT_PLATFORM == target_platform):
                    # IF the current platform exists and is the same as the target platform no need to download a new version
                    module_arch = environment_to_architecture_suffix.get(CURRENT_PLATFORM)
                    log.debug('Target Platform matches Current Platform (%s) for pkg %s ', target_platform, module_name)

                elif download_package_location:
                    # Else there needs to be a provided download location
                    if docker_package_builder.docker_available():
                        log.debug('User Docker to pull pkg %s ', module_name)

                        # Note that the download package function uses a cache so it will only actually pull the first time the user wants to
                        # package this function.
                        rv = docker_package_builder.download_package_and_create_moduleinfo(
                            tmp_distribution_obj,
                            target_platform,
                            module_name,
                            download_package_location
                        )
                        PACKAGE_CACHE[(module_name, original_file_location)] = rv
                        return rv

                    else:
                        raise Exception("Trying to target an architecture with a dependant package but Docker is not available")

                else:
                    raise Exception("Trying to target an architecture with a dependant package but have Docker is not allowed")

            # The package could be either a folder (normal case) or a single python file (ex: 'six' package)
            # If it can not be found as either than there is an issue
            potential_dir = os.path.join(
                pip_packages.get(module_name).location, module_name
            )
            potential_file = os.path.join(
                pip_packages.get(module_name).location, module_name + ".py"
            )

            if os.path.isdir(potential_dir):
                tmp_fp = potential_dir
            elif os.path.isfile(potential_file):
                tmp_fp = potential_file
            else:
                raise Exception

            log.debug("Final fp for %s -> %s", module_name, tmp_fp)

            # Find the dependent modules for this project (distribution obj)
            (
                tmp_dependencies_flat,
                tmp_dependencies_tree,
            ) = _recursive_check_for_dependencies_project(tmp_distribution_obj)


            log.debug("Flat dependencies for %s -> %s", module_name, tmp_dependencies_flat)
            log.debug("Tree dependencies for %s -> %s", module_name, tmp_dependencies_tree)
            
            rv = ModulePackagingInfo(
                **{
                    "module_name": module_name,
                    "type": PackageTypes.PIP,
                    "arch": module_arch,
                    "version_id": tmp_distribution_obj.version,
                    "fp": tmp_fp,
                    "flat": tmp_dependencies_flat,
                    "tree": tmp_dependencies_tree,
                }
            )

        else:
            # Since the module is not in the PIP, platform builtins, or std lib
            # It should be treated as a local package
            mod = sys.modules.get(module_name)
            if mod:
                if not mod.__file__:
                    # if the module __file__ is not present than it is a builtin module to the interpreter
                    tmp_type = PackageTypes.BUILTIN
                    tmp_fp = None
                    tmp_version = None
                else:
                    # this is a local package that was not imported using a relative path
                    tmp_type = PackageTypes.LOCALPACKAGE
                    tmp_version = None

                    if mod.__file__.split("/")[-1] == "__init__.py":
                        tmp_fp = os.path.dirname(mod.__file__)
                    else:
                        tmp_fp = mod.__file__
                    # TODO: Maybe support this in the future
                    log.debug("Local absolute reference for module %s", module_name)
                    raise Exception("Referencing local package using absolute import statement. Change it to a relative import to work with Cdev.")
            elif module_name[0] == ".":
                # IF the module name started with a '.' then it is a relative import
                
                tmp_type = PackageTypes.LOCALPACKAGE
                original_path = Path(original_file_location)

                tmp_version = None

                tmp_module_name = module_name

                levels = 0
                # loop to compute how many layers up this relative import is by popping off the '.' char
                # until it reaches a different char
                while tmp_module_name[0] == ".":
                    tmp_module_name, next_char = tmp_module_name[1:], tmp_module_name[0]

                    if next_char == ".":
                        levels = levels + 1
                    else:
                        break
                
                log.debug("%s if %s levels up from handler", module_name, levels)

                # go up the amount of levels need to get to the top of the search path
                relative_base_dir = original_path.parents[levels - 1]

                # since we popped off the leading '.' chars, the remaining portion is the path to search down from the top
                tmp_pkg_path_parts = tmp_module_name.split(".")

                # Again the module can either be a single file or a directory containing other files
                tmp_potential_file = os.path.join(
                    relative_base_dir,
                    "/".join(tmp_pkg_path_parts[:-1]),
                    tmp_pkg_path_parts[-1] + ".py",
                )
                tmp_potential_dir = os.path.join(
                    relative_base_dir, "/".join(tmp_pkg_path_parts)
                )

                if os.path.isfile(tmp_potential_file):
                    tmp_fp = tmp_potential_file
                    
                elif os.path.isdir(tmp_potential_dir):
                    tmp_fp = tmp_potential_dir
                else:
                    raise Exception

                log.debug("Final relative file location (%s) from workspace to %s", tmp_fp, module_name)
                

            else:
                raise Exception(f"Unmanageable module name for local module {module_name}")

            # Since this is a local package, it does not contain any extra metadata. This means the only way to find the dependencies is to
            # parse each python file in the needed module and look at its imports :/
            (
                dependencies_flat,
                dependencies_tree,
            ) = _recursive_check_for_dependencies_package(tmp_fp)
            log.debug("Flat dependencies for %s -> %s", module_name, dependencies_flat)
            log.debug("Tree dependencies for %s -> %s", module_name, dependencies_tree)

            rv = ModulePackagingInfo(
                **{
                    "module_name": module_name,
                    "type": tmp_type,
                    "version_id": tmp_version,
                    "fp": tmp_fp,
                    "flat": dependencies_flat,
                    "tree": dependencies_tree,
                }
            )

        PACKAGE_CACHE[(module_name, original_file_location)] = rv
        log.debug("Add to cache (%s,%s) => %s", module_name, original_file_location, rv)
        return rv


def _recursive_check_for_dependencies_project(
    project_distribution_obj: pkg_resources.Distribution,
) -> Tuple[List[ModulePackagingInfo], List[ModulePackagingInfo]]:
    """
    Create the ModulePackagingInfo objects for all dependencies of this project. By calling the '_recursive_create_module_package_info' to create the
    ModulePackagingInfo objects, it creates a recursive stack that eventually leads to tmp_flat being a list of ALL needed dependencies and the tmp_tree
    to be the first level in the dependency tree for this project.

    Args:
        project_distribution_obj (pkg_resources.Distribution): object from pkg_resources that contains metadata on this package

    Returns:
        flat (List[ModulePackagingInfo]): List of ALL needed dependencies
        tree (List[ModulePackagingInfo]): First level of a dependency tree for this project
    """
    tmp_flat = set()
    tmp_tree = set()

    for project_obj in project_distribution_obj.requires():

        if project_obj.project_name in PRJ_NAME_TO_TOP_LEVEL_MODULES:
            project_name = project_obj.project_name

        # Note that pip packages should list their dependencies as the project name, but some use the top level module name that
        # can be slightly different than the project name.
        # Example awswrangler (2.12.1) lists 'Requires-Dist: pymysql (>=0.9.0,<1.1.0)' but the actual package name is 'PyMySQL'
        # .....python packaging sucks
        elif project_obj.project_name in MOD_NAME_TO_PRJ_OBJ:
            project_name = MOD_NAME_TO_PRJ_OBJ.get(
                project_obj.project_name
            ).project_name
        else:
            raise Exception

        top_level_modules = PRJ_NAME_TO_TOP_LEVEL_MODULES.get(project_name)

        for req in top_level_modules:

            tmp_dep = _recursive_create_module_package_info(req, None)
            tmp_flat.add(tmp_dep)
            tmp_tree.add(tmp_dep)

            if tmp_dep.flat:
                tmp_flat = tmp_flat.union(set(tmp_dep.flat))

    return tmp_flat, tmp_tree


def _recursive_check_for_dependencies_package(
    fp: Union[FilePath, DirectoryPath]
) -> Tuple[List[ModulePackagingInfo], List[ModulePackagingInfo]]:
    """Create the ModulePackagingInfo objects for all dependencies of a local module. 
    
    By calling the '_recursive_create_module_package_info' to create the ModulePackagingInfo objects,
    it creates a recursive stack that eventually leads to tmp_flat being a list of ALL needed dependencies 
    and the tmp_tree to be the first level in the dependency tree for this project.

    Args:
        fp (Union[FilePath, DirectoryPath]): object from pkg_resources that contains metadata on this package

    Returns:
        flat (List[ModulePackagingInfo]): List of ALL needed dependencies
        tree (List[ModulePackagingInfo]): First level of a dependency tree for this project
    """
    tmp_flat = set()
    tmp_tree = set()

    required_items = _get_local_package_dependencies(fp)

    for req in required_items:
        starting_location = fp if not req[1] else req[1]

        tmp_dep = _recursive_create_module_package_info(req[0], starting_location)

        tmp_flat.add(tmp_dep)
        tmp_tree.add(tmp_dep)

        if tmp_dep.flat:
            tmp_flat = tmp_flat.union(set(tmp_dep.flat))

    return list(tmp_flat), tmp_tree


def _get_local_package_dependencies(
    fp: Union[FilePath, DirectoryPath]
) -> List[Tuple[str, Optional[str]]]:
    """
    Get the local dependencies for a given local module by searching through the file(s) that make up the module and looking
    for import statements.

    Args:
        fp (Union[FilePath, DirectoryPath]): the location of the module in question

    Returns:
        dependencies (List[Tuple[str,str]]): List of dependant module names and an optional filepath if the module is a relative module


    Uses the Serverless Parser library to find all the imports in a given file.
    """
    global _current_working_dir

    if not _current_working_dir:
        # This setting should be set
        raise Exception

    # Since we are just reading the files not importing them. We do not have a guarantee to not cause a circular dependency
    # Therefore, we preload all the files we are going to check before checking them, and add that to a cache to check against
    # If a file is already in the cache, it means it has/or will be accounted for.

    if fp in _already_checked_cache:
        # We have already included this file so skip it
        return []

    # Local Module can be either directory or single file
    if os.path.isdir(fp):
        module_names = set()

        for dir, _, files in os.walk(fp):
            if dir in _already_checked_cache:
                continue

            # Load the directory (and subdirectories)
            _already_checked_cache.add(dir)

            for file in files:
                if not file[-3:] == ".py":
                    continue

                # Load the files
                if os.path.join(dir, file) in _already_checked_cache:
                    continue

                _already_checked_cache.add(os.path.join(dir, file))

                module_names = module_names.union(
                    cdev_parser.parse_file_for_dependencies(os.path.join(dir, file))
                )
                log.debug("Found dependent modules %s for %s", module_names, os.path.join(dir, file))

    else:
        _already_checked_cache.add(fp)
        module_names = cdev_parser.parse_file_for_dependencies(fp)
        
    log.debug("Found dependent modules %s for %s", module_names, fp)

    return list(module_names)







from cdev.constructs.project import Project
from cdev.default.output_manager import CdevOutputManager
from cdev.cli.logger import set_global_logger_from_cli


from core.commands.cloud_output import cloud_output_command as core_cloud_output_command


def cloud_output_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)

    cloud_output_command(config.cloud_output_id)


def cloud_output_command(cloud_output_id: str):

    output_manager = CdevOutputManager()

    myProject = Project.instance()
    
    ws = myProject.get_current_environment().get_workspace()


    core_cloud_output_command(ws, output_manager, cloud_output_id)
    




from typing import Callable, Dict, List, Set


from core.constructs.output_manager import OutputTask

from ..constructs.resource import Resource_Change_Type, Resource_Reference_Difference, Resource_Difference
from ..constructs.mapper import CloudMapper

#from .mappers.aws import aws_lambda, dynamodb, iam, s3, sqs, apigatewayv2, apigateway

from .mappers.simple import (
    api_deployer,
    bucket_deployer,
    lambda_deployer,
    dynamodb_deployer,
    queue_deployer,
    relational_db_deployer,
    static_site_deployer,
    topic_deployer
)

class DefaultMapper(CloudMapper):   
    
    def __init__(self) -> None:
        super().__init__(RESOURCE_TO_HANDLER_FUNCTION)

    def get_namespaces(self) -> List[str]:
        return [
            "cdev::simple::api",
            "cdev::simple::bucket",
            "cdev::simple::function",
            "cdev::simple::lambda_layer",
            "cdev::simple::table",
            "cdev::simple::queue",
            "cdev::simple::relationaldb",
            "cdev::simple::staticsite",
            "cdev::simple::topic"
        ]

    def deploy_resource(self, transaction_token: str, namespace_token: str, resource_diff: Resource_Difference, previous_output: Dict, output_task: OutputTask) -> Dict:
        ruuid = resource_diff.new_resource.ruuid if resource_diff.new_resource else resource_diff.previous_resource.ruuid

        return self.get_resource_to_handler()[ruuid](transaction_token, namespace_token, resource_diff, previous_output, output_task)
    

    def get_available_resources(self) -> Set[str]:
        return set(self.get_namespaces())

    def get_resource_to_handler(self) -> Dict[str, Callable]:
        return self.resource_to_handler
        


RESOURCE_TO_HANDLER_FUNCTION = {
   
    "cdev::simple::function": lambda_deployer.handle_simple_lambda_function_deployment,
    "cdev::simple::lambda_layer": lambda_deployer.handle_simple_layer_deployment,
    "cdev::simple::api": api_deployer.handle_simple_api_deployment,
    "cdev::simple::bucket": bucket_deployer.handle_simple_bucket_deployment,
    "cdev::simple::table": dynamodb_deployer.handle_simple_table_deployment,
    "cdev::simple::queue": queue_deployer.handle_simple_queue_deployment,
    "cdev::simple::relationaldb": relational_db_deployer.handle_simple_relational_db_deployment,
    "cdev::simple::staticsite": static_site_deployer.handle_simple_static_site_deployment,
    "cdev::simple::topic": topic_deployer.handle_simple_topic_deployment
    
}



"""Set of constructs for for making a website to serve Static Web Content

"""
from typing import Any, Optional

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs
from core.constructs.cloud_output import  Cloud_Output_Str
from core.utils import hasher

RUUID = "cdev::simple::staticsite"


class StaticSiteOutput(ResourceOutputs):
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)


    @property
    def bucket_name(self) -> Cloud_Output_Str:
        """The name of the underlying S3 Bucket where the content for the site is stored."""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='bucket_name',
            type=self.OUTPUT_TYPE
        )

    @bucket_name.setter
    def bucket_name(self, value: Any):
        raise Exception


    @property
    def cloudfront_id(self) -> Cloud_Output_Str:
        """The id of the site on the Cloudfront CDN"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='cloudfront_id',
            type=self.OUTPUT_TYPE
        )

    @cloudfront_id.setter
    def cloudfront_id(self, value: Any):
        raise Exception


    @property
    def cloudfront_arn(self) -> Cloud_Output_Str:
        """The arn of the site on the Cloudfront CDN"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='cloudfront_arn',
            type=self.OUTPUT_TYPE
        )

    @cloudfront_arn.setter
    def cloudfront_arn(self, value: Any):
        raise Exception



    @property
    def site_url(self) -> Cloud_Output_Str:
        """The url of the created site"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='site_url',
            type=self.OUTPUT_TYPE
        )

    @site_url.setter
    def site_url(self, value: Any):
        raise Exception


###############
##### Static Site
###############
class simple_static_site_model(ResourceModel):
    index_document: str
    """The suffix for documents when request are made for a folder. ex: site.com/dir1/ will look for /dir1/<index_document>"""
    error_document: str
    """The absolute path of document within the site that will be used as a general error page."""
    sync_folder: bool
    """Whether to consider changes in the state of the content folder to trigger a sync of the content folder."""
    content_folder: Optional[str]
    """The relative path within the Workspace of the folder containting the static content for the site."""
    domain_name: Optional[str]
    """The domain name to associate with this site. This will set up the site to be aliased via DNS to the given domain name."""
    ssl_certificate_arn: Optional[str]
    """Arn of a SSL certificate to use with the site."""

class StaticSite(Resource):
    """A Static Site that can be used to serve static web content. 
    
    """
    
    @update_hash
    def __init__(
        self,
        cdev_name: str,
        index_document: str = "index.html",
        error_document: str = "error.html",
        sync_folder: bool = False,
        content_folder: str = None,
        domain_name: str = None,
        ssl_certificate_arn: str = None,
        nonce: str = ""

    ) -> None:
        """Create a static hosted site.

        Arguments:
            cdev_name (str): [description]
            index_document (str): The suffix for documents when request are made for a folder.
            error_document (str): The absolute path of document within the site that will be used as a general error page.
            sync_folder (bool): Whether to consider changes in the state of the content folder to trigger a sync of the content folder.
            content_folder (str): The relative path within the `Workspace` of the folder containting the static content for the site.
            domain_name (str): The domain name to associate with this site. This will set up the site to be aliased via DNS to the given domain name.
            ssl_certificate_arn (str): Arn of a SSL certificate to use with the site. 
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self._index_document = index_document
        self._error_document = error_document
        self._sync_folder = sync_folder
        self._content_folder = content_folder
        self._domain_name = domain_name
        self._ssl_certificate_arn = ssl_certificate_arn

        self.output = StaticSiteOutput(cdev_name)

    @property
    def index_document(self):
        """The suffix for documents when request are made for a folder."""
        return self._index_document

    @index_document.setter
    @update_hash
    def index_document(self, value: str):
        self._index_document = value

    @property
    def error_document(self):
        """The absolute path of document within the site that will be used as a general error page."""
        return self._error_document

    @error_document.setter
    @update_hash
    def error_document(self, value: str):
        self._error_document = value

    @property
    def content_folder(self):
        """The relative path within the `Workspace` of the folder containting the static content for the site."""
        return self._content_folder

    @content_folder.setter
    @update_hash
    def content_folder(self, value: str):
        self._content_folder = value

    @property
    def sync_folder(self):
        """Whether to consider changes in the state of the content folder to trigger a sync of the content folder."""
        return self._sync_folder

    @sync_folder.setter
    @update_hash
    def sync_folder(self, value: bool):
        self._sync_folder = value

    @property
    def domain_name(self):
        """The domain name to associate with this site. This will set up the site to be aliased via DNS to the given domain name."""
        return self._domain_name

    @domain_name.setter
    @update_hash
    def domain_name(self, value: str):
        self._domain_name = value

    @property
    def ssl_certificate_arn(self):
        """The domain name to associate with this site. This will set up the site to be aliased via DNS to the given domain name."""
        return self._ssl_certificate_arn

    @ssl_certificate_arn.setter
    @update_hash
    def ssl_certificate_arn(self, value: str):
        self._ssl_certificate_arn = value

    def compute_hash(self):
        # TODO update component to look at the content if the sync folder command is given
        self._hash = hasher.hash_list(
            [
                self.index_document,
                self.error_document,
                self.sync_folder,
                self.content_folder,
                self.domain_name,
                self.ssl_certificate_arn,
                self.nonce
            ]
        )

    def render(self) -> simple_static_site_model:
        if self.sync_folder and not self.content_folder:

            raise Exception(f"If sync_folder is set to 'True' then you must provide a path to the folder.")

        return simple_static_site_model(
            ruuid=self.ruuid,
            name=self.name,
            hash=self.hash,
            index_document=self.index_document,
            error_document=self.error_document,
            sync_folder=self.sync_folder,
            content_folder=self.content_folder,  
            domain_name=self.domain_name,
            ssl_certificate_arn=self.ssl_certificate_arn 
        )





"""Utilities for running custom scripts 

"""


from ..constructs.commands import BaseCommand, BaseCommandContainer
from ..constructs.workspace import Workspace
from ..constructs.output_manager import OutputManager


def execute_run_cli(args):
    ws = Workspace.instance()
    
    output_manager = OutputManager()

    run_command(ws, output_manager, args)


def run_command(workspace: Workspace, output: OutputManager, cli_args):
    """
    Attempts to find and run a user defined command

    format:
    cdev run <sub_command> <args>
    """
    # Convert namespace into dict
    params = vars(cli_args)

    # This is the command to run... It can be a single command or a path to the command where the path is '.' delimitated
    sub_command = params.get("subcommand")
    command_args = params.get("args") if params.get("args") else []

    
    try:
        workspace.execute_command(sub_command, command_args)
    except Exception as e:
        print(e)
        return



from core.utils import logger

def set_global_logger_from_cli(log_level: int):
    
    new_log = logger.cdev_logger(
        show_logs=True,
        logging_level=log_level
    )

    logger.set_global_logger(
        new_log
    )

   


# Generated as part of Quick Start project template 

from cdev.resources.simple.xlambda import simple_function_annotation

import aurora_data_api
import pandas

from .utils import helper_function

@simple_function_annotation("hello_world_function")
def hello_world(event, context):
    print('Hello from inside your Function!')
    helper_function()

    return {
        "status_code": 200,
        "message": "Hello Outside World!"
    }


@simple_function_annotation("hello_world_function2")
def hello_world2(event, context):
    print('Hello from inside your Function!')
    print(aurora_data_api)

    return {
        "status_code": 200,
        "message": "Hello Outside World!"
    }


@simple_function_annotation("hello_world_function3")
def hello_world3(event, context):
    print('Hello from inside your Function!')
    print(aurora_data_api)
    print(pandas)

    return {
        "status_code": 200,
        "message": "Hello Outside World!"
    }



@simple_function_annotation("hello_world_function4")
def hello_world4(event, context):

    return {
        "status_code": 200,
        "message": "Hello Outside World!"
    }


from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on the simple table.

put_items: Load data into the table
clear_table: Clear all data from the table

"""



"""Set of constructs for making Serverless based Relational DBs

"""

from enum import Enum
from typing import Any

from core.constructs.resource import Resource, ResourceModel, ResourceOutputs, update_hash, PermissionsAvailableMixin
from core.constructs.cloud_output import  Cloud_Output_Str, OutputType
from core.utils import hasher

from core.default.resources.simple.iam import Permission, PermissionArn


RUUID = "cdev::simple::relationaldb"


#####################
###### Permission
######################
class RelationalDBPermissions:
     
    def __init__(self, resource_name: str) -> None:
        
        self.DATABASE_ACCESS = Permission(
            actions=[
                "rds-data:BatchExecuteStatement",
                "rds-data:BeginTransaction",
                "rds-data:CommitTransaction",
                "rds-data:ExecuteStatement",
                "rds-data:RollbackTransaction",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Access to the DB"""

        self.SECRET_ACCESS = PermissionArn(
            arn="arn:aws:iam::aws:policy/SecretsManagerReadWrite"
        )
        """Access to the generated Secret that contains the connection information"""


##############
##### Output
##############
class RelationalDBOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a Relational DB has been deployed."""   
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)

    @property
    def cluster_arn(self) -> Cloud_Output_Str:
        """The name of the generated db cluster"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='cluster_arn',
            type=self.OUTPUT_TYPE
        )

    @cluster_arn.setter
    def cluster_arn(self, value: Any):
        raise Exception

    @property
    def endpoint(self) -> Cloud_Output_Str:
        """The name of the generated db cluster"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='endpoint',
            type=self.OUTPUT_TYPE
        )

    @endpoint.setter
    def endpoint(self, value: Any):
        raise Exception

    @property
    def secret_arn(self) -> Cloud_Output_Str:
        """The arn of the secret value create that holds the password for the db"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='secret_arn',
            type=self.OUTPUT_TYPE
        )

    @secret_arn.setter
    def secret_arn(self, value: Any):
        raise Exception


###############
##### RelationalDB
###############
class db_engine(str, Enum):
    aurora = "aurora"
    """MySQL 5.6-compatible Aurora"""

    aurora_mysql = "aurora-mysql"
    """MySQL 5.7-compatible Aurora"""

    aurora_postgresql = "aurora-postgresql"
    """Postgres 10.4-compatible Aurora"""



class simple_relational_db_model(ResourceModel):
    """Model that represents a relation db"""
    Engine: db_engine
    """DB engine"""
    MasterUsername: str
    """Username used to connect to the DB"""
    MasterUserPassword: str
    """Username used to connect to the DB"""
    DatabaseName: str
    """Name for the main db"""
    EnableHttpEndpoint: bool
    """Allow connection via a generated HTTP endpoint"""
    MaxCapacity: int
    """Maximum amount of capacity to scale to"""
    MinCapacity: int    
    """Amount of seconds to wait before scaling DB completely down."""
    SecondsToPause: int
    """Amount of seconds to wait before scaling DB completely down"""



class RelationalDB(PermissionsAvailableMixin, Resource):
    """Construct for creating a Serverless Relational DB"""
    
    @update_hash
    def __init__(
        self,
        cdev_name: str,
        engine: db_engine,
        username: str,
        password: str,
        database_name: str = "",
        enable_http_endpoint: bool = True,
        max_capacity: int = 64,
        min_capacity: int = 2,
        seconds_to_pause: int = 300,
        nonce: str = ""
    ) -> None:
        """
        Args:
            cdev_name (str): Name of the resource
            engine (db_engine): DB engine
            username (str): Username used to connect to the DB
            password (str): Password used to connect to the DB
            database_name (str, optional): Name for the main db. Defaults to "".
            enable_http_endpoint (bool, optional): Allow connection via a generated HTTP endpoint. Defaults to True.
            max_capacity (int, optional): Maximum amount of capacity to scale to. Defaults to 64.
            min_capacity (int, optional): Minimum amount of capacity to scale to. Defaults to 2.
            seconds_to_pause (int, optional): Amount of seconds to wait before scaling DB completely down. Defaults to 300.
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self._engine = engine.value
        self._master_username = username
        self._master_user_password = password
        self._database_name = database_name
        self._enable_http_endpoint = enable_http_endpoint
        self._max_capacity = max_capacity
        self._min_capacity = min_capacity
        self._seconds_to_pause = seconds_to_pause

        self.output: RelationalDBOutput = RelationalDBOutput(cdev_name)
        self.available_permissions: RelationalDBPermissions = RelationalDBPermissions(cdev_name)

    @property
    def seconds_to_pause(self):
        """Amount of seconds to wait before scaling DB completely down"""
        return self._seconds_to_pause

    @seconds_to_pause.setter
    @update_hash
    def seconds_to_pause(self, value: int):
        self._seconds_to_pause = value

    @property
    def min_capacity(self):
        """Amount of seconds to wait before scaling DB completely down."""
        return self._min_capacity

    @min_capacity.setter
    @update_hash
    def min_capacity(self, value: int):
        self._min_capacity = value

    @property
    def max_capacity(self):
        """Maximum amount of capacity to scale to"""
        return self._max_capacity

    @max_capacity.setter
    @update_hash
    def max_capacity(self, value: int):
        self._max_capacity = value

    @property
    def enable_http_endpoint(self):
        """Allow connection via a generated HTTP endpoint"""
        return self._enable_http_endpoint

    @enable_http_endpoint.setter
    @update_hash
    def enable_http_endpoint(self, value: bool):
        self._enable_http_endpoint = value

    @property
    def database_name(self):
        """Name for the main db"""
        return self._database_name

    @database_name.setter
    @update_hash
    def database_name(self, value: str):
        self._database_name = value

    @property
    def master_user_password(self):
        """Password used to connect to the DB"""
        return self._master_user_password

    @master_user_password.setter
    @update_hash
    def master_user_password(self, value: str):
        self._master_user_password = value

    @property
    def master_username(self):
        """Username used to connect to the DB"""
        return self._master_username

    @master_username.setter
    @update_hash
    def master_username(self, value: str):
        self._master_username = value

    @property
    def engine(self):
        """DB engine"""
        return self._engine

    @engine.setter
    @update_hash
    def engine(self, value: db_engine):
        self._engine = value

    def compute_hash(self):
        self._hash = hasher.hash_list(
            [
                self.engine,
                self.master_username,
                self.master_user_password,
                self.database_name,
                self.enable_http_endpoint,
                self.max_capacity,
                self.min_capacity,
                self.seconds_to_pause,
                self.nonce
            ]
        )

    def render(self) -> simple_relational_db_model:
        return simple_relational_db_model(
            **{
                "ruuid": self.ruuid,
                "name": self.name,
                "hash": self.hash,
                "Engine": self.engine,
                "MasterUsername": self.master_username,
                "MasterUserPassword": self.master_user_password,
                "DatabaseName": self.database_name,
                "EnableHttpEndpoint": self.enable_http_endpoint,
                "MaxCapacity": self.max_capacity,
                "MinCapacity": self.min_capacity,
                "SecondsToPause": self.seconds_to_pause,
            }
        )




"""Base Exceptions to use within the framework

These are the two exceptions that are expected to be used at the module level throughout Cdev. Within a module, it should have its own more precise
exceptions, but they should be wrapped in these classes before being exposed at the module level. Any other error that is raised will be treated as 
an uncaught exception and not the responsibility of other modules. 

Warning: Something is not correct but the process can continue running.
Error: Something has reached a state that should terminate the process. 

"""
import sys


class Cdev_Warning(Exception):
    def __init__(self, msg: str) -> None:
        self.msg = msg
        super().__init__(self.msg)


class Cdev_Error(Exception):
    def __init__(self, msg: str, nested_exceptions: Exception=None) -> None:
        self.msg = msg
        self.nested_exceptions = nested_exceptions
        super().__init__(self.msg)


def end_process():
    sys.exit(1)



# Generated as part of the Slack Events project template 
import json
import os

from slack_sdk import signature
from slack_sdk import WebClient

from cdev.resources.simple.api import Api, route_verb
from cdev.resources.simple.xlambda import simple_function_annotation

from cdev import Project as cdev_project

from ..project_settings import SlackBotSettings

myProject = cdev_project.instance()

mySettings: SlackBotSettings =  myProject.settings

DemoApi = Api("demoapi")

webhook_route = DemoApi.route("/webhook", route_verb.POST)


env_vars = {
    "SLACK_SECRET": mySettings.SLACK_SECRET, 
    "SLACK_BOT_OAUTH_TOKEN": mySettings.SLACK_BOT_OAUTH_TOKEN,
}


signature_verifier = signature.SignatureVerifier(signing_secret=os.environ.get('SLACK_SECRET'))
client = WebClient(token=os.environ.get('SLACK_BOT_OAUTH_TOKEN'))

@simple_function_annotation("webhook", events=[webhook_route.event()], environment=env_vars)
def webhook(event, context):
    # Load the info to validate the request
    body = event.get('body')
    timestamp = event.get('headers').get("x-slack-request-timestamp")
    slack_signature = event.get('headers').get("x-slack-signature")

    is_valid = signature_verifier.is_valid(body, timestamp, slack_signature)

    if not is_valid:
        # Not a valid request from our Slack App so return 401
        return {
            "status_code": 401,
        }


    data = json.loads(event.get("body"))
    print(data)

    if data.get("type") == "url_verification":
        return {
            "status_code": 200,
            "message": {
                "challenge": data.get('challenge')
            }
        }

    response_channel = data.get('event').get('channel')

    client.chat_meMessage(
        channel=response_channel,
        text="Hello from your app! :tada:",
    )
    
    return {
        "status_code": 200,
    }


myProject.display_output("Base API URL", DemoApi.output.endpoint)
myProject.display_output("Routes", DemoApi.output.endpoints)


"""Utilities for working with user defined commands

[detailed description]

"""
import importlib
import inspect
import os
import sys
from typing import List, Tuple, Union

from core.constructs.commands import BaseCommand, BaseCommandContainer
from core.utils.module_loader import import_module


class AmbiguousCommandName(Exception):
    pass


class NoCommandFound(Exception):
    pass


class TooManyCommandClasses(Exception):
    pass


def find_specified_command(
    command_list: List[str], 
    all_search_locations_list: List[str]
) -> Tuple[Union[BaseCommand, BaseCommandContainer], str, str, bool]:
    
    command_list_copy = command_list.copy()

    for search_location in all_search_locations_list:
        # Search locations are denoted as python module paths, so convert the module to a path to search from
        mod = import_module(search_location)
        search_location_path = os.path.dirname(mod.__file__)
    

        did_find, is_command = _recursive_find_specified_command(
            command_list.copy(), search_location_path
        )

        if not did_find:
            continue

        full_command_module_name = f"{search_location}.{'.'.join(command_list_copy)}"

        if is_command:
            try:
                initialized_object = initialize_command_module(
                    full_command_module_name
                )
            except Exception as e:
                print(e)
                raise Exception(f"ERROR Initializing Command Module at {full_command_module_name}")

        else:
            try:
                initialized_object = initialize_command_container_module(
                    full_command_module_name
                )
            except Exception as e:
                print(e)
                raise Exception(f"ERROR Initializing Command Container Module at {search_location}.{command_list}")

        
        final_path_name =  f"{search_location}.{'.'.join(command_list_copy[:-1])}"
        command_name = command_list_copy[-1]
        return initialized_object, final_path_name, command_name, is_command


    # If we loop through all the search locations and no commands were found then throw exception
    raise NoCommandFound(f"No Command or Commands Container found for {command_list_copy}")
        

def initialize_command_module(mod_path: str) -> BaseCommand:
    
    mod = import_module(mod_path, denote_output=True)

    # Check for the class that derives from BaseCommand... if there is more then one class then throw error (note this is a current implementation detail)
    # because it is easier if their is only one command per file so that we can use the file name as the command name
    _has_found_a_valid_command = False
    initialized_obj = None
    for item in dir(mod):
        potential_obj = getattr(mod, item)
        if (
            inspect.isclass(potential_obj)
            and issubclass(potential_obj, BaseCommand)
            and not (potential_obj == BaseCommand)
        ):
            if _has_found_a_valid_command:
                raise TooManyCommandClasses

            _has_found_a_valid_command = True

            # initialize an instance of the class
            initialized_obj = potential_obj()

    if not initialized_obj:
        raise NoCommandFound

    return initialized_obj


def initialize_command_container_module(mod_path: str) -> BaseCommandContainer:
    mod = import_module(mod_path)
    

    # Check for the class that derives from BaseCommandContainer... if there is more then one class then throw error (note this is a current implementation detail)
    # because it is easier if their is only one command per file so that we can use the file name as the command name
    _has_found_a_valid_command_container = False
    initialized_obj = None
    for item in dir(mod):
        potential_obj = getattr(mod, item)
        if (
            inspect.isclass(potential_obj)
            and issubclass(potential_obj, BaseCommandContainer)
            and not (potential_obj == BaseCommandContainer)
        ):
            if _has_found_a_valid_command_container:
                # TODO better exception
                return

            _has_found_a_valid_command_container = True

            initialized_obj = potential_obj()

    if not initialized_obj:
        raise NoCommandFound(f"Could not find command Container")

    
    return initialized_obj


def _recursive_find_specified_command(
    command_list: List[str], search_path: str
) -> Tuple[bool, bool]:
    """Return if the given command list results in a potential found command or command container

    Recursively check at each step if the next part in the command list is valid. If at the end of the 
    list return that something was found. If at any point there is a missing link then return False.

    Args:
        command_list (List[str]): List of command parts
        search_path (str): Path the search from

    Returns:
        Tuple[bool, bool]: Found, Is_Command
    """
    

    if len(command_list) == 1:
        # A location was specified and we are the final part of the command so the command must be a py file in this dir
        # of be a directory that has a command container
        if os.path.isfile(os.path.join(search_path, command_list[0]+".py")):
            return (True, True)

        if os.path.isdir(os.path.join(search_path, command_list[0])):
            if os.path.join(os.path.join(search_path, command_list[0], "__init__.py")):
                return (True, False)

        return (False, False)


    next_command_part = command_list.pop(0)

    if not next_command_part in os.listdir(search_path):
        return (False, False)

    return _recursive_find_specified_command(command_list, os.path.join(search_path, next_command_part))





"""Set of constructs for adding permissions and authorization 

"""

from typing import Dict, FrozenSet, List, Optional, Union

from core.constructs.models import ImmutableModel
from core.constructs.cloud_output import Cloud_Output_Str
from core.constructs.types import cdev_str_model, cdev_str
from core.utils import hasher


class permission_model(ImmutableModel):
    actions: FrozenSet[str]
    cloud_id: cdev_str_model
    effect: str
    resource_suffix: Optional[str]
    hash: str

    class Config:
        use_enum_values = True
        # Beta Feature but should be fine since this is simple data 
        frozen = True


class permission_arn_model(ImmutableModel):
    arn: str
    hash: str
    
    class Config:
        use_enum_values = True
        # Beta Feature but should be fine since this is simple data 
        frozen = True


class Permission():
    """
    Permission that can be attached to a resource to give it permission to access other resources.
    """

    def __init__(
        self,
        actions: List[str],
        cloud_id: cdev_str,
        effect: str,
        resource_suffix: Optional[str] = "",
    ):
        """
        Arguments:
            actions (List[str]): List of the actions that this policy will include
            cloud_id (cdev_str): The cloud id of the resource that is giving the permission
            effect ('Allow', 'Deny'): Allow or Deny the permission
            resource_suffix (Optional[str]): Some permissions need suffixes added to the looked up aws resource (i.e. dynamodb streams )
        """
        self.actions = actions
        self.cloud_id = cloud_id
        self.effect = effect
        self.resource_suffix = resource_suffix
            
    def render(self) -> permission_model:

        return permission_model(
            actions=frozenset(self.actions),
            cloud_id=self.cloud_id.render() if isinstance(self.cloud_id, Cloud_Output_Str) else self.cloud_id,
            effect=self.effect,
            hash=self.hash()
        )

    def hash(self) -> str:
        _hash = hasher.hash_list([
            hasher.hash_list(self.actions),
            self.cloud_id.hash() if isinstance(self.cloud_id, Cloud_Output_Str) else self.cloud_id,
            self.effect
        ])

        return _hash


class PermissionArn():
    """
    Id of a permission that is already deployed on the cloud.
    """
    def __init__(self, arn: str) -> None:
        self.arn = arn


    def render(self) -> permission_arn_model:
        return permission_arn_model(
            arn=self.arn
        )

    def hash(self) -> str:
        return hasher.hash_string(self.arn)




from typing import Dict, List

from core.constructs.resource import ResourceModel
from core.constructs.workspace import Workspace

RUUID = "cdev::simple::table"

def get_cloud_output_from_cdev_name(component_name: str, cdev_name: str) -> str:
    try:
        ws = Workspace.instance()


        cloud_output = ws.get_backend().get_cloud_output_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return cloud_output
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None



def get_resource_from_cdev_name(component_name: str, cdev_name: str) -> ResourceModel:
    try:
        ws = Workspace.instance()


        resource = ws.get_backend().get_resource_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return resource
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None




_attribute_types_to_python_type = {"S": [str], "N": [int, float], "B": [bytes]}


def validate_data(data: Dict, attributes: Dict, keys: List[Dict]) -> bool:
    for key in keys:
        if not key.get("attribute_name") in data:
            return (False, f"Table key {key.get('attribute_name')} not in data")

        valid_types = _attribute_types_to_python_type.get(
            attributes.get(key.get("attribute_name"))
        )

        is_val_valid_type = False
        for attribute_type in valid_types:
            if isinstance(data.get(key.get("attribute_name")), attribute_type):
                is_val_valid_type = True
                break

        if not is_val_valid_type:
            return (
                False,
                f"Table key {key.get('attribute_name')} ({valid_types}) not correct type in data ({type(data.get(key.get('attribute_name')))})",
            )

    return (True, "")


def recursive_translate_data(value) -> Dict:
    if isinstance(value, str):
        transformed_val = {"S": value}
    elif isinstance(value, (int, float)):
        transformed_val = {"N": value}
    elif isinstance(value, float):
        transformed_val = {"N": value}
    elif isinstance(value, bytes):
        transformed_val = {"B": value}
    elif isinstance(value, bool):
        transformed_val = {"BOOl": value}
    elif not value:
        transformed_val = {"NULL": True}
    elif isinstance(value, list):
        if all(isinstance(x, str) for x in value):
            transformed_val = {"SS": [x for x in value]}
        elif all(isinstance(x, (int, float)) for x in value):
            transformed_val = {"NS": [x for x in value]}
        elif all(isinstance(x, bytes) for x in value):
            transformed_val = {"BS": [x for x in value]}
        else:
            transformed_val = {"L": [recursive_translate_data(value)]}

    elif isinstance(value, dict):
        transformed_val = {k: recursive_translate_data(v) for k, v in value.items()}

    else:
        raise Exception

    return transformed_val



import docker
import json
import os
from pkg_resources import Distribution
import re
from typing import List

from core.constructs.workspace import Workspace


from .utils import (
    PackageTypes,
    ModulePackagingInfo,
    lambda_python_environment,
    parse_requirement_line,
    CONTAINER_NAMES,
    environment_to_architecture_suffix
)

from core.utils.logger import log


def docker_available() -> bool:
    return True

has_run_container = False
build_container = None


class DockerDownloadCache:
    """
    Naive cache implentation for know which files have been downloaded.
    """

    def __init__(self, cache_location: str) -> None:
        self._cache_location = os.path.join(cache_location, "cache.json")
        self._downloads_locations = os.path.join(cache_location, "downloads")

        if not os.path.isdir(self._downloads_locations):
            os.mkdir(self._downloads_locations)


        if not os.path.isfile(self._cache_location):
            self._cache = {
                lambda_python_environment.py37: {},
                lambda_python_environment.py38_x86_64: {},
                lambda_python_environment.py38_arm64: {},
                lambda_python_environment.py39_x86_64: {},
                lambda_python_environment.py39_arm64: {},
                lambda_python_environment.py3_x86_64: {},
                lambda_python_environment.py3_arm64: {},
            }

        else:
            with open(self._cache_location) as fh:
                self._cache = json.load(fh)

        self._cache_dirs = {
            lambda_python_environment.py37: "py37",
            lambda_python_environment.py38_x86_64: "py38x8664",
            lambda_python_environment.py38_arm64: "py38arm64",
            lambda_python_environment.py39_x86_64: "py39x8664",
            lambda_python_environment.py39_arm64: "py39arm64",
            lambda_python_environment.py3_x86_64: "py3x8664",
            lambda_python_environment.py3_arm64: "py3arm64",
        }

    def find_item(self, environment: lambda_python_environment, id: str):
        raw_data = self._cache.get(environment).get(id)
        if raw_data:
            return [ModulePackagingInfo(**x) for x in raw_data]
        else:
            None

    def add_item(
        self,
        environment: lambda_python_environment,
        id: str,
        item: ModulePackagingInfo,
    ):
        self._cache[environment][id] = item

        with open(self._cache_location, "w") as fh:
            json.dump(self._cache, fh, indent=4)

    def get_packaging_dir(self, environment: lambda_python_environment):
        FULL_DIR = os.path.join(self._downloads_locations, self._cache_dirs.get(environment))

        if not os.path.isdir(FULL_DIR):
            os.mkdir(FULL_DIR)

        return FULL_DIR

DOWNLOAD_CACHE:  DockerDownloadCache = None 


def download_package_and_create_moduleinfo(
    project: Distribution, 
    environment: lambda_python_environment, 
    module_name: str,
    downloads_directory: str
) -> ModulePackagingInfo:
    """Download a project in a platform compatible way to extract a particular top level module info from the project. 
    
    Note that this function implements a cache so that it only downloads the project the first time it ever is called 
    for the pair (project_name, environment). This means that most calls to this function will not have to actually download the project.

    Args:
        project (Distribution): The distribution object that contains the metadata for the package
        environment (lambda_python_environments): Deployment environment the project needs to support
        module_name (str): The top level module from this project that we want
        base_archive_directory (str): Path to Cache
    Returns:
        info (ModulePackagingInfo): Object for the information on packaging this module
    """
    global DOWNLOAD_CACHE

    DOWNLOAD_CACHE = DockerDownloadCache(downloads_directory)

    
    package_information = _download_package(project, environment)

    potential_modules = [x for x in package_information if x.module_name == module_name]

    if not len(potential_modules) == 1:
        raise Exception

    final_module = potential_modules[0]

    return final_module


def _download_package(
    project: Distribution, environment: lambda_python_environment
) -> List[ModulePackagingInfo]:
    """
    Perform the actual downloading of the package and then parse out the needed information about the top level modules made available
    from the project. Uses the cache to make sure that way we only download the project when actually needed. Also reuses the same container
    when downloading packages so that it uses pip's cache to not redownload the same package.

    Args:
        project (Distribution): The distribution object that contains the metadata for the package
        environment (lambda_python_environments): Deployment environment the project needs to support

    Returns:
        info (List[ModulePackagingInfo]): ModulePackagingInfo for all the top level modules in the project

    """
    global has_run_container
    global build_container

    cache_item = DOWNLOAD_CACHE.find_item(environment, project.project_name)
    if cache_item:
        log.debug("Hit Cache for Download Pkg (%s, %s)", environment, project.project_name)
        return cache_item

    packaging_dir = DOWNLOAD_CACHE.get_packaging_dir(environment)

    
    client = docker.from_env()

    if not environment in CONTAINER_NAMES:
        raise Exception(f"No available Images for building projects on environment {environment}")

    image_name = CONTAINER_NAMES.get(environment)

    print(f"PULLING DOCKER IMAGE {image_name}")

    try:
        client.images.pull(image_name)
    except Exception as e:
        print(e)
        raise Exception(f"Could Not Pull Docker Images {image_name}")
        
    print(f"PULLED IMAGE")

    try:
        if not has_run_container:
            build_container = client.containers.run(
                image_name,
                entrypoint="/var/lang/bin/pip",
                command=f"install {project.project_name}=={project.version} --target /tmp",
                volumes=[f"{packaging_dir}:/tmp"],
                detach=True,
            )

            has_run_container = True

        else:
            build_container.restart()

            build_container.exec_run(
                cmd=f"/var/lang/bin/pip install {project.project_name}=={project.version} --target /tmp",
                detach=True,
            )

        for x in build_container.logs(stream=True):
            msg = x.decode("ascii")
            print(f"Container Output -> {msg}")

    except Exception as e:
        print(e)
        raise Exception(f"Could Install {project.project_name} ({project.version}) using Image {image_name}")

    # Create the actual packaging information and store it in cache via this function
    info = _create_package_info(project.project_name, environment)

    return info


def _create_package_info(
    project_name: str, environment: lambda_python_environment
) -> List[ModulePackagingInfo]:
    """Creates a list of ModulePackagingInfo objects that represent the top level modules made available from this
    package. 
    
    It uses the information from the 'dist-info' that was downloaded for the projects by PIP. This function
    recursively calls itself to compute the information for dependant projects that were downloaded.

    Arg:
        project_name (str): Name of the project to create the information for
        environment (lambda_python_environments): Deployment environment the project needs to support

    Returns:
        info (List[ModulePackagingInfo]): ModulePackagingInfo objects for each of the top level modules in this project
    """

    # We check and add to the cache at this layer because we recursively call this function to compute the
    # information about dependencies of a project. Since projects can share dependencies it saves time from recomputing
    # this info
    cache_item = DOWNLOAD_CACHE.find_item(environment, project_name)
    if cache_item:
        log.debug("Hit Cache for Download Pkg (%s, %s)", environment, project_name)
        return cache_item

    packaging_dir = DOWNLOAD_CACHE.get_packaging_dir(environment)

    dist_info_dir = None

    for _, dir_names, _ in os.walk(packaging_dir):
        # We need to look through the packaging dir to find the dist-info folder for the project.
        # Note the dist-info folder is <project_name>-<version>.dist-info
        # https://www.python.org/dev/peps/pep-0376/#one-dist-info-directory-per-installed-distribution
        regex_dist_info = "(.*).dist-info"

        for dir_name in dir_names:
            m = re.match(regex_dist_info, dir_name)
            if not m:
                continue

            # [<project_name>-<version>, dist-info]
            dir_name_split = m.groups()

            # project_name
            split_name_version = dir_name_split[0].split("-")

            name = split_name_version[0]
            version = split_name_version[1]

            # Note that if a project contains '-' they are converted to '_' so that the above regex/string parsing works, so convert any '-'
            # into '_' when looking for the correct directory
            if name == project_name.replace("-", "_"):
                dist_info_dir = os.path.join(
                    packaging_dir, f"{name}-{version}.dist-info"
                )
                log.debug("Found dist info for (%s, %s) at %s", environment, project_name, dist_info_dir)
                break

        break

    if not dist_info_dir:
        raise Exception(f"Could not find dist info for {project_name} -> {dist_info_dir}")

    if not os.path.isdir(os.path.join(packaging_dir, dist_info_dir)):
        raise Exception(f"COULD NOT FIND {os.path.join(packaging_dir, dist_info_dir)}")

    # If no top level file is found then assume the only top level module is the same as the project name
    if not os.path.isfile(os.path.join(packaging_dir, dist_info_dir, "top_level.txt")):
        top_level_module_names = [project_name]

    else:
        with open(os.path.join(packaging_dir, dist_info_dir, "top_level.txt")) as fh:
            top_level_module_names = [x.strip() for x in fh.readlines()]

        log.debug("top level modules for (%s, %s) are %s", environment, project_name, top_level_module_names)

    tmp_flat_requirements: List[ModulePackagingInfo] = []
    tmp_tree_requirements: List[ModulePackagingInfo] = []
    rv = []

    if not os.path.isfile(os.path.join(packaging_dir, dist_info_dir, "METADATA")):
        # If not metadata file is found then assume no dependencies
        log.debug("COULD NOT FIND %s", os.path.join(packaging_dir, dist_info_dir, 'METADATA'))

    else:
        required_packages = set()
        current_package_version = ""
        with open(os.path.join(packaging_dir, dist_info_dir, "METADATA")) as fh:
            lines = fh.readlines()

            for line in lines:

                if not line:
                    break

                if line.split(":")[0] == "Version":
                    # https://www.python.org/dev/peps/pep-0345/#version
                    # ex:
                    # version: 1.0.2
                    current_package_version = line.split(":")[1].strip()

                if line.split(":")[0] == "Requires-Dist":
                    # https://www.python.org/dev/peps/pep-0345/#requires-dist-multiple-use
                    # ex:
                    # Requires-Dist: pytz (>=2017.3)
                    stripped_information = line.split(":")[1].strip()


                    requirement_project_name = parse_requirement_line(
                        stripped_information
                    )

                    if requirement_project_name:
                        log.debug("Pkg %s requires pkg %s", project_name, requirement_project_name)
                        required_packages.add(requirement_project_name)

        for required_package in required_packages:
            # Recursively find the requirements to each needed project 
            dependent_pkg_infos = _create_package_info(required_package, environment)
            for dependent_pkg_info in dependent_pkg_infos:
                # For the flat data structure add both the dependency and the dependencies dependencies 
                tmp_flat_requirements.extend(dependent_pkg_info.flat)
                tmp_flat_requirements.append(dependent_pkg_info)

                tmp_tree_requirements.append(dependent_pkg_info)

    for top_level_module_name in top_level_module_names:
        # Create module packaging info objects for each of the top level modules of the pkg

        # The package could be either a folder (normal case) or a single python file (ex: 'six' package)
        # If it can not be found as either than there is an issue
        potential_dir = os.path.join(packaging_dir, top_level_module_name)
        potential_file = os.path.join(packaging_dir, top_level_module_name + ".py")

        if os.path.isdir(potential_dir):
            tmp_fp = potential_dir

        elif os.path.isfile(potential_file):
            tmp_fp = potential_file

        else:
            # just continue cause things like numpy have __dummy__ in their top level pkg names but dont provide it
            continue

        info = ModulePackagingInfo(
            **{
                "module_name": top_level_module_name,
                "type": PackageTypes.PIP,
                "version_id": current_package_version,
                "arch": environment_to_architecture_suffix.get(environment),
                "fp": tmp_fp,
                "flat": tmp_flat_requirements,
                "tree": tmp_tree_requirements,
            }
        )

        rv.append(info)

    # Add the information to the cache for this function
    DOWNLOAD_CACHE.add_item(environment, project_name, [x.dict() for x in rv])

    return rv



from core.constructs.resource import ResourceModel
from core.constructs.workspace import Workspace

RUUID = "cdev::simple::table"

def get_cloud_output_from_cdev_name(component_name: str, cdev_name: str) -> str:
    try:
        ws = Workspace.instance()


        cloud_output = ws.get_backend().get_cloud_output_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return cloud_output
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None



def get_resource_from_cdev_name(component_name: str, cdev_name: str) -> ResourceModel:
    try:
        ws = Workspace.instance()


        resource = ws.get_backend().get_resource_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_name
        )

        return resource
    except Exception as e:
        print(f"Could not find resource {component_name}:{RUUID}:{cdev_name}")
        print(e)
        return None




from typing import Any, Dict
from uuid import uuid4


from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.constructs.output_manager import OutputTask
from core.default.resources.simple import table
from core.utils import hasher

from .. import aws_client 

def _create_simple_dynamodb_table(
    transaction_token: str, 
    namespace_token: str, 
    resource: table.simple_table_model, 
    output_task: OutputTask
) -> bool:

    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])
    table_name = f"cdev-table-{full_namespace_suffix}"

    output_task.update(comment='[blink]Creating Table. This will take a few seconds.[/blink]')

    rv = aws_client.run_client_function(
        "dynamodb",
        "create_table",
        {
            "TableName": table_name,
            "AttributeDefinitions": [{
                "AttributeName": x.attribute_name,
                "AttributeType": x.attribute_type.value
            } for x in resource.attributes],
            "KeySchema": [{
                "AttributeName": x.attribute_name,
                "KeyType": x.key_type.value
            } for x in resource.keys],
            "BillingMode": "PAY_PER_REQUEST",
        },
        wait={"name": "table_exists", "args": {"TableName": table_name}},
    )

    output_info = {
        "table_name": table_name,
        "cloud_id": rv.get("TableDescription").get("TableArn"),
        "arn": rv.get("TableDescription").get("TableArn"),
        "cdev_name": resource.name,
        "ruuid": resource.ruuid,
    }

    return output_info


def _update_simple_dynamodb_table(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: table.simple_table_model,
    new_resource: table.simple_table_model,
    previous_output: Dict,
    output_task: OutputTask
    ) -> bool:
    raise Exception



def _remove_simple_dynamodb_table(
    transaction_token: str,  
    previous_output: Dict, 
    output_task: OutputTask
    ) -> bool:

    table_name = previous_output.get("table_name")

    output_task.update(comment='[blink]Deleting Table. This will take a few seconds.[/blink]')

    aws_client.run_client_function(
        "dynamodb",
        "delete_table",
        {
            "TableName": table_name,
        },
        wait={"name": "table_not_exists", "args": {"TableName": table_name}},
    )



def handle_simple_table_deployment(transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
        ) -> Dict:
    
    if resource_diff.action_type == Resource_Change_Type.CREATE:
        return _create_simple_dynamodb_table(
            transaction_token,
            namespace_token,
            resource_diff.new_resource,
            output_task
        )
    elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:

        return _update_simple_dynamodb_table(
            transaction_token,
            namespace_token,
            table.simple_table_model(**resource_diff.previous_resource.dict()),
            resource_diff.new_resource,
            previous_output,
            output_task
        )
        
    elif resource_diff.action_type == Resource_Change_Type.DELETE:

        _remove_simple_dynamodb_table(
            transaction_token,
            previous_output,
            output_task
        )

        return {}



from enum import Enum
import platform
import sys


CURRENT_PLATFORM = platform.machine() if platform.machine() in ['x86_64', 'aarch64'] else None


class lambda_python_environment(str, Enum):
    py37 = "py37"
    py38_x86_64 = "py38-x86_64"
    py38_arm64 = "py38-arm64"
    py39_x86_64 = "py39-x86_64"
    py39_arm64 = "py39-arm64"
    py3_x86_64 = "py3-x86_64"
    py3_arm64 = "py3-arm64"


def get_current_closest_platform() -> lambda_python_environment:
    python_version = f'{sys.version_info[0]}_{sys.version_info[1]}'


    if python_version == "3_7":
        return lambda_python_environment.py37


    elif python_version == "3_8":
        if platform.machine() == 'x86_64':
            return lambda_python_environment.py38_x86_64

        elif platform.machine() == 'aarch64' or platform.machine() == 'arm64' or platform.machine().startswith('arm'):
            return lambda_python_environment.py38_arm64

        else:
            # TODO Raise warning instead that then notifies the user they need to set the default platform
            print(f"Could not find directly compatible deployment platform for your platform ({platform.machine()}). Some third party dependencies might not work correctly.")
            return None


    elif python_version == "3_9":
        if platform.machine() == 'x86_64':
            return lambda_python_environment.py39_x86_64

        elif platform.machine() == 'aarch64' or platform.machine() == 'arm64' or platform.machine().startswith('arm'):
            return lambda_python_environment.py39_arm64

        else:
            # TODO Raise warning instead that then notifies the user they need to set the default platform
            print(f"Could not find directly compatible deployment platform for your platform ({platform.machine()}). Some third party dependencies might not work correctly.")
            return None



from typing import List, Dict

from pydantic import BaseModel
from pydantic.types import FilePath

from core.constructs.workspace import Workspace, Workspace_Info




class environment_info(BaseModel):
    """
    Represents the information about an environment.

    Arguments:
        name (str): Name of this environment
        workspace_info (Workspace_Info): The information needed to load the Workspace for this environment
    """

    name: str
    workspace_info: Workspace_Info

    def __init__(
        __pydantic_self__,
        name: str,
        workspace_info: Workspace_Info
        ) -> None:

        super().__init__(
            **{"name": name, "workspace_info": workspace_info}
        )


class Environment:
    """
    A logically isolated instance of a project.
    """

    def __init__(self, info: environment_info) -> None:
        pass

    def get_workspace(self) -> Workspace:
        raise NotImplementedError

    def initialize_environment(self):
        raise NotImplementedError




from argparse import ArgumentParser
import json
import os
from typing import List, Dict

from boto3 import client

from core.constructs.commands import BaseCommand, OutputWrapper
from core.constructs.workspace import Workspace
from core.utils import hasher



from core.default.commands.function.utils import get_cloud_id_from_cdev_name

RUUID = "cdev::simple::function"


class execute(BaseCommand):

    help = """
        Execute a function in the cloud.
    """

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "function_id", 
            type=str, 
            help="The id of the function to execute."
        )
        parser.add_argument(
            "--event",
            type=str,
            help="File (json) location of event object to provide as input to the function. Can not be used with '--event-data` flag.",
        )
        parser.add_argument(
            "--event-data",
            type=str,
            help="Raw string form of event object to provide as input to the function. Can not be used with '--event' flag.",
        )
        
       

    def command(self, *args, **kwargs):

        full_function_name: str = kwargs.get("function_id")

        event_file_location: str = kwargs.get("event")
        event_raw_data: str = kwargs.get("event_data")

        if event_file_location and event_raw_data:
            raise Exception("Can not provide both '--event-data' and '--event'")

        event_data = {}

        if event_file_location:
            if not os.path.isfile(event_file_location):
                raise Exception(f"{event_file_location} is not a valid file location")

            with open(event_file_location) as fh:
                try:
                    event_data = json.load(fh)
                except Exception as e:
                    print(e)
                    raise Exception(f'Could not load {event_file_location} as json')

                
        if event_raw_data:
            try:
                event_data = json.loads(event_raw_data)
            except Exception as e:
                print(e)
                raise Exception(f'Could not load {event_raw_data} as json')


        component_name = full_function_name.split('.')[0]
        function_name = full_function_name.split('.')[1]

        cloud_name = get_cloud_id_from_cdev_name(component_name, function_name)

        lambda_client = client('lambda')    

        self.stdout.write(f'executing {full_function_name}')
        response = lambda_client.invoke(
            FunctionName=cloud_name,
            InvocationType='RequestResponse',
            Payload=json.dumps(event_data)
        )


        print(response)


"""Structures for representing cloud resources and references within the framework

"""
from functools import wraps
from typing import Dict, TypeVar, Union, List, Optional, Set, Callable, Any

from enum import Enum

from core.constructs.models import ImmutableModel
from core.constructs.cloud_output import OutputType, Cloud_Output_Str
from core.constructs.types import F

from ..utils.hasher import hash_list

    
##################
##### Resource
##################
class ResourceModel(ImmutableModel):
    """Base class with basic information about a defined resource.
    
    All classes derived from this should contain additional information about the state of a resource. 
    Within the life cycle of the Cdev Core framework,this is used to represent resources after any
    configuration; hence the need to make these immutable data
    classes. 

    Attributes:
        name: The name of the resource. Note that the combination of a name and ruuid should be unique 
          within a namespace
        ruuid: The identifier of this resource type.
        hash: A string that is the hash of this object. This hash must be computed such that it changes 
          only if a change in the state is desired.
    """

    name: str
    """
    This is a human readable logical name for the resource. This must be unique for the resource within the namespace and resource type
    (i.e. the concat value of <ruuid>:<name> must be unique)

    This value is important for allow human level refactoring of resources. To update a resources name once created, you must edit only the 
    name and not change the hash. If you change both the hash and name in the same deployment, it will register this as a delete and create 
    instead of update. 
    """

    ruuid: str
    """
    Name space identifier that is used to pass this resource to a downstream mapper

    Form: (top-level-namespace)::(resource-type-id)
    """

    hash: str
    """
    This is a hash that is used to identify if changes in the resources have occurred.
    """

    class Config:
        arbitrary_types_allowed = True


class Resource_Change_Type(str, Enum):
    CREATE = "CREATE"
    UPDATE_IDENTITY = "UPDATE_IDENTITY"
    UPDATE_NAME = "UPDATE_NAME"
    DELETE = "DELETE"


class Resource_Difference(ImmutableModel):
    action_type: Resource_Change_Type
    component_name: str
    previous_resource: Optional[ResourceModel]
    new_resource: Optional[ResourceModel]

    def __init__(
        __pydantic_self__,
        action_type: Resource_Change_Type,
        component_name: str,
        previous_resource: ResourceModel = None,
        new_resource: ResourceModel = None,
    ) -> None:
        super().__init__(
            **{
                "action_type": action_type,
                "component_name": component_name,
                "previous_resource": previous_resource,
                "new_resource": new_resource,
            }
        )

    class Config:
        use_enum_values = True
        # Beta Feature but should be fine since this is simple data 
        frozen = True


def update_hash(func: F) -> F:
    """Wrap a function that modifies the state of a Resource so that the hash is properly updated when a change occurs.

    Args:
        func (F): State modifying method.

    Returns:
        F: Wrapped function to compute new hash after changes have been made
    """
    @wraps(func)
    def wrapped_func(resource: "Resource", *func_posargs, **func_kwargs):
        rv = func(resource, *func_posargs, **func_kwargs)
        resource.compute_hash()
        return rv

    return wrapped_func


class Resource:
    def __init__(self, name: str, ruuid: str, nonce: str = ""):
        self._name = name
        self._ruuid = ruuid
        self._hash = None
        self._nonce = nonce

    def render(self) -> ResourceModel:
        raise NotImplementedError

    @property
    def name(self):
        """The name of the created resource"""
        return self._name

    @name.setter
    def name(self, value: str):
        raise Exception("The Name of a resource can only be set when initialized.")

    @property
    def ruuid(self):
        """The ruuid of the created resource"""
        return self._ruuid

    @ruuid.setter
    def ruuid(self, value: str):
        raise Exception("The Ruuid of a resource can only be set when initialized.")

    @property
    def nonce(self):
        """The nonce of the created resource"""
        return self._nonce

    @nonce.setter
    def nonce(self, value: str):
        raise Exception("The nonce of a resource can only be set when initialized.")

    @property
    def hash(self):
        """The cdev hash of the resource"""
        if self._hash is None:
            raise Exception

        return self._hash

    @hash.setter
    def hash(self, value: str):
        raise Exception("The `hash` of a Resource can only be set by the `compute_hash` method")

    def compute_hash(self):
        raise NotImplementedError


class ResourceOutputs():
    """Container object for the returned values from the cloud after the resource has been deployed."""
    OUTPUT_TYPE = OutputType.RESOURCE

    def __init__(self, name: str, ruuid: str) -> None:
        self._name = name
        self._ruuid = ruuid

    @property
    def cloud_id(self) -> 'Cloud_Output_Str':
        return Cloud_Output_Str(
            name=self._name,
            ruuid=self._ruuid,
            key='cloud_id',
            type=self.OUTPUT_TYPE
        )

    @cloud_id.setter
    def cloud_id(self, value: Any):
        raise Exception

    @property
    def cloud_region(self) -> 'Cloud_Output_Str':
        return Cloud_Output_Str(
            name=self._name,
            ruuid=self._ruuid,
            key='cloud_region',
            type=self.OUTPUT_TYPE
        )

    @cloud_region.setter
    def cloud_region(self, value: Any):
        raise Exception

##################
##### Reference
##################

class ResourceReferenceModel(ImmutableModel):
    """
    This is the information needed to reference a resource that is defined outside this component

    --- Attributes ---

    - ruuid ->  a string that represents the resource identifier (provider::resource-type)

    - hash  ->  a string that is the hash of this object. This hash must be computed such that
                it changes only if a change in the state is desired.
    """

    component_name: str
    """
    Name of the component that this resources resides in
    """

    ruuid: str
    """
    Name space identifier that is used to pass this resource to a downstream mapper

    Form: (top-level-namespace)::(resource-type-id)
    """

    name: str
    """
    This is a human readable logical name for the resource. This must be unique for the resource within the namespace and resource type
    (i.e. the concat value of <ruuid>:<name> must be unique)

    This value is important for allow human level refactoring of resources. To update a resources name once created, you must edit only the 
    name and not change the hash. If you change both the hash and name in the same deployment, it will register this as a delete and create 
    instead of update. 
    """

    is_in_parent_resource_state: Optional[bool]
    """
    Boolean to determine if the reference should be resolved in the same resource state or from the parent resource state.
    """

    def __init__(
        __pydantic_self__,
        component_name: str,
        ruuid: str,
        name: str,
        is_in_parent_resource_state: bool = False,
    ) -> None:
        super().__init__(
            **{
                "component_name": component_name,
                "ruuid": ruuid,
                "name": name,
                "is_in_parent_resource_state": is_in_parent_resource_state,
            }
        )

    class Config:
        extra = "allow"
        frozen = True


class Resource_Reference_Change_Type(str, Enum):
    CREATE = "CREATE"
    DELETE = "DELETE"


class Resource_Reference_Difference(ImmutableModel):
    action_type: Resource_Reference_Change_Type
    originating_component_name: str
    resource_reference: ResourceReferenceModel

    def __init__(
        __pydantic_self__,
        action_type: Resource_Reference_Change_Type,
        originating_component_name: str,
        resource_reference: ResourceReferenceModel,
    ) -> None:
        super().__init__(
            **{
                "action_type": action_type,
                "originating_component_name": originating_component_name,
                "resource_reference": resource_reference,
            }
        )

    class Config:
        use_enum_values = True
        frozen = True


class Resource_Reference:
    RUUID: str = None

    def __init__(
        self, component_name: str, name: str, is_in_parent_resource_state: bool = False
    ):
        self.component_name = component_name
        self.name = name
        self.is_in_parent_resource_state = is_in_parent_resource_state

    def render(self) -> ResourceReferenceModel:
        raise NotImplementedError

    def compute_hash(self) -> str:
        return hash_list(
            [
                self.component_name,
                self.RUUID,
                self.name,
                str(self.is_in_parent_resource_state),
            ]
        )



####################
##### Mixins
####################
class PermissionsAvailableMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # forwards all unused arguments
        self._available_permissions = None # Needs to be set later in the calling subclass

    @property
    def available_permissions(self):
        """The available permissions of the created resource"""
        return self._available_permissions

    @available_permissions.setter
    def available_permissions(self, permissions: Any):
        self._available_permissions = permissions



class PermissionsGrantableMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # forwards all unused arguments
        self._granted_permissions = [] # Needs to be set later in the calling subclass


    @property
    def granted_permissions(self):
        return self._granted_permissions

    @granted_permissions.setter
    @update_hash
    def granted_permissions(self, value: List):
        self._granted_permissions = value



# Generated as part of Quick Start project template
import os

from cdev.default.cloudmapper import DefaultMapper
from cdev.default.components import Cdev_FileSystem_Component

from cdev import Project as cdev_project

myProject = cdev_project.instance()


myProject.add_mapper(
    DefaultMapper()
)


myProject.add_component(
    Cdev_FileSystem_Component(os.path.join("src","hello_world"), "hello_world_comp")
)





from cdev.constructs.project import Project
from cdev.default.output_manager import CdevOutputManager
from cdev.cli.logger import set_global_logger_from_cli
from rich.prompt import Confirm


from core.constructs.workspace import Workspace_State

def destroy_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)
    destroy_command(args)


def destroy_command(args):

    output_manager = CdevOutputManager()
    myProject = Project.instance()

    ws = myProject.get_current_environment().get_workspace()


    ws.set_state(Workspace_State.EXECUTING_FRONTEND)

    diff_previous_component_names = [x.name for x in ws.get_backend().get_resource_state(ws.get_resource_state_uuid()).components]


    differences = ws.create_state_differences([], diff_previous_component_names)

    if any(x for x in differences):
        output_manager.print_state_differences(differences)


    

    print("")
    do_deployment = Confirm.ask("Are you sure you want to Destroy the environment?")

    if not do_deployment:
        return

    differences_structured = ws.sort_differences(differences)
    ws.set_state(Workspace_State.EXECUTING_BACKEND)

    ws.deploy_differences(differences_structured)



"""Command for syncing a set of resources to a static site
"""
from argparse import ArgumentParser
import boto3
import os
import mimetypes

from core.constructs.commands import BaseCommand, OutputWrapper
from core.default.resources.simple.object_store import Bucket, bucket_model
from core.utils.paths import get_full_path_from_workspace_base

from . import utils


class sync_files(BaseCommand):
    help = """
    Command to sync a folder of content to a bucket
    """

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "resource_name", type=str, help="The bucket resource you want to sync."
        )
        parser.add_argument(
            "--dir",
            type=str,
            help="content folder to sync to the resource.",
        )
        parser.add_argument(
            "--clear",
            action="store_true",
            help="Clear the existing content of the bucket before syncing the new data.",
        )

    def command(self, *args, **kwargs):
        full_resource_name: str = kwargs.get("resource_name")
        component_name = full_resource_name.split('.')[0]
        bucket_cdev_name = full_resource_name.split('.')[1]
        content_directory = kwargs.get("dir")
        clear_bucket = kwargs.get("clear")

        cloud_output = utils.get_cloud_output_from_cdev_name(component_name, bucket_cdev_name)

        final_dir = get_full_path_from_workspace_base(content_directory)

        bucket_name = cloud_output.get("bucket_name")
        

        s3 = boto3.resource("s3")
        bucket = s3.Bucket(bucket_name)

        if clear_bucket:
            bucket.object_versions.delete()


        for subdir, dirs, files in os.walk(final_dir):
            for file in files:
                full_path = os.path.join(subdir, file)

                key_name = os.path.relpath(full_path, final_dir)

                mimetype, _ = mimetypes.guess_type(full_path)
                if mimetype is None:
                    raise Exception("Failed to guess mimetype")
                self.stdout.write(f"Uploading file -> {full_path}")
                bucket.upload_file(
                    full_path, key_name, ExtraArgs={"ContentType": mimetype}
                )



"""Utilities for creating a representation of a desired state of the workspace

"""
from typing import List, Tuple

from ..constructs.workspace import Workspace, Workspace_State
from ..constructs.components import Component_Difference
from ..constructs.resource import Resource_Difference, Resource_Reference_Difference
from ..constructs.output_manager import OutputManager

from core.utils.logger import log


def execute_frontend_cli(args):

    WORKSPACE = Workspace.instance()

    output_manager = OutputManager()


    
    execute_frontend(WORKSPACE, output_manager)



def execute_frontend(workspace: Workspace, output: OutputManager, previous_component_names: List[str]=None) -> Tuple[List[Component_Difference], List[Resource_Difference], List[Resource_Reference_Difference]]:

    log.debug("Executing Frontend")

    workspace.set_state(Workspace_State.EXECUTING_FRONTEND)
    current_state = workspace.generate_current_state()


    output.print_local_state(current_state)

    if not previous_component_names:
        diff_previous_component_names = [x.name for x in workspace.get_backend().get_resource_state(workspace.get_resource_state_uuid()).components]
    else:
        diff_previous_component_names = previous_component_names

    
    output.print_components_to_diff_against(diff_previous_component_names)

    differences = workspace.create_state_differences(current_state, diff_previous_component_names)


    if any(x for x in differences):
        output.print_state_differences(differences)

    else:
        print("No Differences")
        return None

    log.debug("Finish Executing Frontend")
    
    return differences



"""Extra types to be used within the type hints of the framework
"""
from typing import  TypeVar, Callable, Any

from core.constructs.cloud_output import Cloud_Output_Bool, Cloud_Output_Int, Cloud_Output_Str, cloud_output_dynamic_model


cdev_str = TypeVar('cdev_str', str, Cloud_Output_Str)
cdev_str_model = TypeVar('cdev_str_model', str, cloud_output_dynamic_model)

F = TypeVar("F", bound=Callable[..., Any])



from ..constructs.workspace import Workspace
from ..constructs.output_manager import OutputManager

from core.utils.logger import log

def cloud_output_command_cli(args):
    config = args[0]
    cloud_output_command(config)


def cloud_output_command(workspace: Workspace, output: OutputManager, cloud_output_id: str):
    log.debug("Executing Frontend")

   

    split_names = cloud_output_id.split('.')

    if not len(split_names) == 4:
        raise Exception("Output not provided in correct structure. ex: <component>.<ruuid>.<cdev_name>.<output_key>")


    component_name = split_names[0]
    resource_ruuid = f"cdev::simple::{split_names[1]}"
    resource_name = split_names[2]
    output_key = split_names[3]


    try:
        cloud_output = workspace.get_backend().get_cloud_output_by_name(
            workspace.get_resource_state_uuid(),
            component_name,
            resource_ruuid,
            resource_name
        )
    except Exception as e:
        print(e)
        raise e


    if not output_key in cloud_output:
        raise Exception(f"Key {output_key} not in Cloud Output {cloud_output}")


    output._console.print(f"[blue]{output_key}[/blue] -> {cloud_output.get(output_key)}")


from cdev.constructs.project import Project
from cdev.cli.logger import set_global_logger_from_cli
from cdev.default.output_manager import CdevOutputManager

from core.commands.execute_frontend import execute_frontend
from core.utils.logger import log



def plan_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)

    plan_command({})


def plan_command(args):
    log.info(msg="Starting Plan Command")
    output_manager = CdevOutputManager()

    myProject = Project.instance()
    log.debug("Loaded Project Global Instance")
    
    ws = myProject.get_current_environment().get_workspace()

    
    
    execute_frontend(ws, output_manager)
    log.info("Finished Plan Command")

  



"""
Basic available settings 
"""
import os

from typing import Any, Optional
from rich.logging import RichHandler

from rich.traceback import install

from typing import List, Set

from pydantic import (
    BaseModel,
    BaseSettings,
)

from core.utils.module_loader import import_class, import_module


class Settings_Info(BaseModel):
    base_class: str
    user_setting_module: Optional[List[str]]
    secret_dir: Optional[str]
    

class Settings(BaseSettings):

    # Starting Path for workspace
    BASE_PATH: str = os.path.abspath(os.getcwd())


    INTERNAL_FOLDER_NAME: str = ".cdev"

    # Configuration file for a workspace
    WORKSPACE_FILE: str = "workspace_info.json"


    INTERMEDIATE_FOLDER_LOCATION: str = os.path.join(
        BASE_PATH, INTERNAL_FOLDER_NAME, "intermediate"
    )

    # Bucket to use as a place to store resource artifacts in the cloud
    S3_ARTIFACTS_BUCKET: str = "cdev-demo-project-artifacts"

    # AWS account information
    AWS_REGION: str = "us-east-1"
    
    # Base entry point file for the workspace
    ENTRY_POINT_FILE: str = os.path.join(
        BASE_PATH, "cdev_project.py"
    )

    DEPLOYMENT_PLATFORM: str = "x86"

    USE_DOCKER: bool = False


    class Config:
        env_prefix = 'cdev_'
        validate_assignment = True
        #extra = 'allow'



def initialize_settings(info: Settings_Info) -> Settings:
    class_name = info.base_class.split('.')[-1]
    module_name = ".".join(info.base_class.split('.')[:-1])

    base_settings_class = import_class(module_name, class_name)

    kw_args = {}

    if info.secret_dir:
        kw_args['_secrets_dir'] = info.secret_dir


    # All the settings are stored as relative paths so they need to convert to full paths
    t = {k:os.path.join(os.getcwd(),v) for k,v in kw_args.items()}
    
    base_setting_obj = base_settings_class(**t)

    if info.user_setting_module:
        for settings_module in info.user_setting_module:
            

            module = import_module(settings_module)

            for setting in dir(module):
                if setting.isupper():
                    setting_value = getattr(module, setting)

                    setattr(base_setting_obj, setting, setting_value)

       

    return base_setting_obj

    





from time import sleep
from typing import Dict, List
from uuid import uuid4

from core.default.resources.simple import xlambda as simple_lambda
from core.default.resources.simple import api as simple_api
from core.default.resources.simple import object_store as simple_bucket
from core.default.resources.simple import table as simple_table
from core.default.resources.simple import queue as simple_queue
from core.default.resources.simple import topic as simple_topic

from .. import aws_client

from . import bucket_deployer
from . import topic_deployer


##############################################
###### API GATEWAY ROUTE EVENTS
##############################################
def _handle_adding_api_event(event: simple_api.route_event_model, cloud_function_id: str) -> Dict:
    # Needed info
    #   - Event model
    #   - originating resource output
    #   - responding resource output
    
    # Returns:
    #   - Info about the deployed stuff

    api_id = event.api_id
    route_id = event.route_id
    
    args = {
        "IntegrationType": "AWS_PROXY",
        "IntegrationUri": cloud_function_id,
        "PayloadFormatVersion": "2.0",
        "IntegrationMethod": event.verb,
        "ApiId": api_id,
    }
    
    integration_rv = aws_client.run_client_function(
        "apigatewayv2", "create_integration", args
    )


    # Now that the integration has been created we need to attach it to the apigateway route
    update_info = {
        "ApiId": api_id,
        "RouteId": route_id,
        "Target": f"integrations/{integration_rv.get('IntegrationId')}",
    }

    aws_client.run_client_function("apigatewayv2", "update_route", update_info)

    aws_region = aws_client.get_aws_region()
    aws_account = aws_client.get_account_number()

    # Add permission to lambda to allow apigateway to invoke this function
    stmt_id = f"stmt-{route_id}"
    permission_model_args = {
        "FunctionName": cloud_function_id,
        "Action": "lambda:InvokeFunction",
        "Principal": "apigateway.amazonaws.com",
        "StatementId": stmt_id,
        "SourceArn": f"arn:aws:execute-api:{aws_region}:{aws_account}:{api_id}/*/{event.verb}{event.path}",
    }

    aws_client.run_client_function(
        "lambda", "add_permission", permission_model_args
    )

    return {
        "integration_id": integration_rv.get("IntegrationId"),
        "permission_stmt_id": stmt_id,
        "api_id": api_id,
        "route_id": route_id
    }


def _handle_deleting_api_event(event: dict, function_cloud_id: str) -> bool:
   
    cloud_id = function_cloud_id
    integration_id = event.get("integration_id")
    stmt_id = event.get("permission_stmt_id")
    
    api_id = event.get("api_id")
    route_id = event.get("route_id")

    # Delete the permission on the lambda function
    aws_client.run_client_function(
        "lambda",
        "remove_permission",
        {
            "FunctionName": cloud_id, 
            "StatementId": stmt_id
        },
    )

    # To delete a route event from a simple api we need to delete the integration.    
    # Leave the target blank so that the route has no integration.
    # Then delete the actual integration cloud obj
    update_info = {"ApiId": api_id, "RouteId": route_id, "Target": ""}

    aws_client.run_client_function("apigatewayv2", "update_route", update_info)


    aws_client.run_client_function(
        "apigatewayv2",
        "delete_integration",
        {
            "ApiId": api_id, 
            "IntegrationId": integration_id
        },
    )

##############################################
##### DYNAMODB TABLE STREAM EVENT
##############################################


def _handle_adding_stream_event(event: simple_table.stream_event_model, cloud_function_id) -> Dict:
    table_name = event.table_name
    view_type = event.view_type
    batch_size = event.batch_size

    rv = aws_client.run_client_function(
        "dynamodb", 
        "describe_table", 
        {
            "TableName": table_name
        }
    )

    if not rv.get("Table").get("StreamSpecification"):
        rv = aws_client.run_client_function(
            "dynamodb",
            "update_table",
            {
                "TableName": table_name,
                "StreamSpecification": {
                    "StreamEnabled": True,
                    "StreamViewType": view_type,
                },
            },
        )
        table_data = rv.get("TableDescription")

    else:
        if not rv.get("Table").get("StreamSpecification").get("StreamEnabled"):
            rv = aws_client.run_client_function(
                "dynamodb",
                "update_table",
                {
                    "TableName": table_name,
                    "StreamSpecification": {
                        "StreamEnabled": True,
                        "StreamViewType": view_type,
                    },
                },
            )
            table_data = rv.get("TableDescription")

        else:
            table_data = rv.get("Table")

    stream_arn = table_data.get("LatestStreamArn")

    sleep(5)
    rv = aws_client.run_client_function(
        "lambda",
        "create_event_source_mapping",
        {
            "EventSourceArn": stream_arn,
            "FunctionName": cloud_function_id,
            "Enabled": True,
            "BatchSize": batch_size,
            "StartingPosition": "LATEST",
        },
    )

    uuid = rv.get("UUID")

    return {
        "stream_arn": stream_arn,  
        "stream_event_id": uuid
    }


def _handle_deleting_stream_event(event: dict, function_cloud_id: str):
    uuid = event.get("stream_event_id")

    aws_client.run_client_function(
        "lambda", "delete_event_source_mapping", {"UUID": uuid}
    )



##############################################
##### BUCKET EVENT TRIGGER
##############################################



def _handle_adding_bucket_event(bucket_event: simple_bucket.bucket_event_model , cloud_function_id: str) -> Dict:

    bucket_arn = bucket_event.bucket_arn
    bucket_name = bucket_event.bucket_name
    events = [bucket_event.bucket_event_type.value]

    # Add permission to lambda to allow s3 to invoke this function
    stmt_id = f"stmt-{str(uuid4())}"
    permission_model_args = {
        "FunctionName": cloud_function_id,
        "Action": "lambda:InvokeFunction",
        "Principal": "s3.amazonaws.com",
        "StatementId": stmt_id,
        "SourceArn": bucket_arn,
    }

    aws_client.run_client_function(
        "lambda", "add_permission", permission_model_args
    )

    # Add trigger to the bucket... use helper function in the bucket deployer because bucket can send events to sqs and sns also
    print(f"event to add {events}")

    # Wait a few seconds for the newly added permissions to take hold
    # if this happens too fast, it will say the lambda does not have correct permissions to be triggered. 
    sleep(5)
    
    bucket_event_id = bucket_deployer.add_eventsource(
        bucket_name,
        bucket_deployer.event_hander_types.LAMBDA,
        cloud_function_id,
        events
    )

    return {
        "bucket_event_id": bucket_event_id,
        "permission_stmt_id": stmt_id,
        "bucket_name": bucket_name,
    }


def _handle_deleting_bucket_event(event: dict, function_cloud_id: str):
    
    bucket_event_id = event.get("bucket_event_id")
    bucket_name =event.get("bucket_name")
    stmt_id = event.get("permission_stmt_id")


    # Delete the permission on the lambda function
    aws_client.run_client_function(
        "lambda",
        "remove_permission",
        {
            "FunctionName": function_cloud_id, 
            "StatementId": stmt_id
        },
    )

    bucket_deployer.remove_eventsource(bucket_name, bucket_event_id)


##############################################
##### QUEUE EVENT TRIGGER
##############################################


def _handle_adding_queue_event(event: simple_queue.queue_event_model, cloud_function_id) -> Dict:
    queue_arn = event.queue_arn
    batch_size = event.batch_size

    rv = aws_client.run_client_function(
        "lambda",
        "create_event_source_mapping",
        {
            "EventSourceArn": queue_arn,
            "FunctionName": cloud_function_id,
            "Enabled": True,
            "BatchSize": batch_size,
        },
    )

    uuid = rv.get("UUID")

    return {
        "queue_event_id": uuid
    }


def _handle_deleting_queue_event(event: dict, function_cloud_id: str):
    queue_event_id = event.get('queue_event_id')

    aws_client.run_client_function(
        "lambda", 
        "delete_event_source_mapping",
        {
            "UUID": queue_event_id
        }
    )



##############################################
##### TOPIC EVENT TRIGGER
##############################################


def _handle_adding_topic_subscription(event: simple_topic.topic_event_model, cloud_function_id) -> Dict:
    topic_arn = event.topic_arn

    # Add permission to lambda to allow apigateway to invoke this function
    stmt_id = f"stmt-{str(uuid4())}"
    permission_model_args = {
        "FunctionName": cloud_function_id,
        "Action": "lambda:InvokeFunction",
        "Principal": "sns.amazonaws.com",
        "StatementId": stmt_id,
        "SourceArn": topic_arn,
    }

    aws_client.run_client_function(
        "lambda", "add_permission", permission_model_args
    )

    subscribe_arn = topic_deployer.add_subscriber(
        topic_arn, "lambda", cloud_function_id
    )

    return {
        "topic_event_id": subscribe_arn, 
        "permission_stmt_id": stmt_id
    }


def _handle_deleting_topic_subscription(event: dict, function_cloud_id: str):
    subscription_arn = event.get("topic_event_id")
    permission_stmt_id = event.get("permission_stmt_id")

    # Delete the permission on the lambda function
    aws_client.run_client_function(
        "lambda",
        "remove_permission",
        {
            "FunctionName": function_cloud_id, 
            "StatementId": permission_stmt_id
        },
    )

    # Remove subscription
    topic_deployer.remove_subscriber(subscription_arn)

    return True


EVENT_TO_HANDLERS = {
    simple_lambda.RUUID: {
        simple_api.RUUID:{   
            "CREATE": _handle_adding_api_event,
            "REMOVE": _handle_deleting_api_event,
        },
        simple_bucket.RUUID:{
            "CREATE": _handle_adding_bucket_event,
            "REMOVE": _handle_deleting_bucket_event,
        },
        simple_table.RUUID:{
            "CREATE": _handle_adding_stream_event,
            "REMOVE": _handle_deleting_stream_event,
        },
        simple_queue.RUUID:{
            "CREATE": _handle_adding_queue_event,
            "REMOVE": _handle_deleting_queue_event,
        },
        simple_topic.RUUID:{
            "CREATE": _handle_adding_topic_subscription,
            "REMOVE": _handle_deleting_topic_subscription,
        }
    },
    
}



"""Set of constructs for making fan our message topics

"""
from typing import Any

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs, PermissionsAvailableMixin
from core.constructs.cloud_output import Cloud_Output_Str, OutputType
from core.constructs.types import cdev_str_model

from core.default.resources.simple.events import Event, event_model
from core.default.resources.simple.iam import Permission

from core.utils import hasher

RUUID = "cdev::simple::topic"


######################
##### Events
######################
class topic_event_model(event_model):
    """Model representing a Topic event"""
    topic_arn: cdev_str_model   


class TopicEvent(Event):
    """Construct representing the Events from the Topic"""
    
    def __init__(self, topic_name: str) -> None:

        self.cdev_topic_name = topic_name

        self.topic_arn = Cloud_Output_Str(
                name=topic_name, 
                ruuid=RUUID, 
                key='arn',
                type=OutputType.RESOURCE 
            )


    def render(self) -> topic_event_model:
        return topic_event_model(
            originating_resource_name=self.cdev_topic_name,
            originating_resource_type=RUUID,
            hash=self.hash(),
            topic_arn=self.topic_arn.render(),
        )


    def hash(self) -> str:
        return hasher.hash_list([
            self.cdev_topic_name,
        ])


#####################
###### Permission
######################
class TopicPermissions:
    def __init__(self, resource_name) -> None:
        self.SUBSCRIBE = Permission(
            actions=[
                "sns:GetTopicAttributes",
                "sns:Subscribe",
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to subscribe to the Topic"""

        self.PUBLISH = Permission(
            actions=[
                "sns:GetTopicAttributes",
                "sns:Publish"
            ],
            cloud_id=Cloud_Output_Str(resource_name, RUUID, 'cloud_id', OutputType.RESOURCE),
            effect="Allow",
        )
        """Permission to publish to the Topic"""

##############
##### Output
##############
class TopicOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a Topic has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)

    @property
    def topic_name(self) -> Cloud_Output_Str:
        """The name of the generated topic"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='topic_name',
            type=self.OUTPUT_TYPE
        )

    @topic_name.setter
    def topic_name(self, value: Any):
        raise Exception


###############
##### Topic
###############
class simple_topic_model(ResourceModel):
    """Model representing a Pub/Sub Topic"""

    is_fifo: bool
    """Should the Topic guarantee ordering of messages"""

class Topic(PermissionsAvailableMixin, Resource):
    """Construct for creating a managed SNS Pub/Sub Topic"""

    @update_hash
    def __init__(
        self, cdev_name: str, is_fifo: bool = False, nonce: str = ''
    ) -> None:
        """
        Args:
            cdev_name (str): Name of the resource.
            is_fifo (bool, default=False): If True, the Queue will guarantee ordering of messages.
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self._is_fifo = is_fifo
        self.output = TopicOutput(cdev_name)
        self.available_permissions: TopicPermissions = TopicPermissions(cdev_name)
        self._event = None

    @property
    def is_fifo(self):
        """Should the Pub/Sub Topic guarantee ordering of messages"""
        return self._is_fifo

    @is_fifo.setter
    @update_hash
    def is_fifo(self, value: bool):
        self._is_fifo = value

    def create_event_trigger(
        self
    ) -> TopicEvent:
        """Create an Event for the Topic that other resources can listen to

        Raises:
            Exception: _description_
            Exception: _description_

        Returns:
            QueueEvent
        """
        if self._event:
            raise Exception("Already created an event on this topic. Use `get_event()` to get the current event.")
    

        event = TopicEvent(
            topic_name=self.name,
        )

        self._event = event

        return event

    def get_event(self) -> TopicEvent:
        """Get the Event for this Queue

        Raises:
            Exception: _description_

        Returns:
            QueueEvent
        """
        if not self._event:
            raise Exception("Topic Event has not been created. Create a Topic Event for this topic using the `create_event_trigger` function before calling this function.")

        return self._event

    def compute_hash(self):
        self._hash = hasher.hash_list([self.is_fifo, self.nonce])

    def render(self) -> simple_topic_model:
        return simple_topic_model(
            name=self.name,
            ruuid=self.ruuid,
            hash=self.hash,
            is_fifo=self.is_fifo,
        )

